"""
Fixtures and helper functions to test models and test resources
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *
from collections import namedtuple

from sqlalchemy import ForeignKey
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative.api import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.sqltypes import Integer, String

from zsl.application.modules.alchemy_module import SessionHolder
from zsl.db.model.app_model import AppModel
from zsl.db.model.raw_model import ModelBase
from zsl.utils.injection_helper import inject

Base = declarative_base()


class UserAppModel(AppModel):
    pass


class UserModel(Base, ModelBase):
    __tablename__ = 'users'
    __app_model__ = UserAppModel

    id = Column('id', Integer, primary_key=True, nullable=False)
    name = Column('name', String, nullable=False)

    def __repr__(self):
        return '<User(id=\'{}\',name=\'{}\')>'.format(self.id, self.name)

UserTuple = namedtuple('UserTuple', 'id name')

users = [
    UserTuple(1, 'one'),
    UserTuple(2, 'two'),
    UserTuple(3, 'three'),
    UserTuple(4, 'four'),
    UserTuple(5, 'five'),
    UserTuple(6, 'six'),
    UserTuple(7, 'seven'),
    UserTuple(8, 'eight'),
    UserTuple(9, 'nine'),
    UserTuple(10, 'ten'),
    UserTuple(11, 'one'),
    UserTuple(12, 'two'),
    UserTuple(13, 'three'),
    UserTuple(14, 'four'),
    UserTuple(15, 'five'),
    UserTuple(16, 'six'),
    UserTuple(17, 'seven'),
    UserTuple(18, 'eight'),
    UserTuple(19, 'nine'),
    UserTuple(20, 'ten')
]


class AddressAppModel(AppModel):
    pass


class AddressModel(Base, ModelBase):
    __tablename__ = 'addresses'
    __app_model__ = AddressAppModel

    id = Column(Integer, primary_key=True)
    email_address = Column(String, nullable=False)
    user_id = Column(Integer, ForeignKey('users.id'))
    user = relationship('UserModel', back_populates='addresses')

    def __repr__(self):
        return '<Address(email_address=\'{}\')>'.format(self.email_address)

UserModel.addresses = relationship('AddressModel', order_by=AddressModel.id, back_populates='user')


AddressTuple = namedtuple('AddressTuple', 'id email_address user_id')

addresses = [
    AddressTuple(1, 'knutley0@reverbnation.com', 1),
    AddressTuple(2, 'zlerwell1@aboutads.info', 1),
    AddressTuple(3, 'ereap2@deviantart.com', 2),
]


def get_non_existent_id():
    """Return an id for non existent user."""
    return len(users) + 10


@inject(engine=Engine, session_holder=SessionHolder)
def create_resource_test_data(engine, session_holder):
    Base.metadata.create_all(engine)
    session = session_holder()

    for u in users:
        session.add(UserModel(**u._asdict()))

    for a in addresses:
        session.add(AddressModel(**a._asdict()))

    session.commit()
    session.close()



"""
:mod:`zsl.utils.email_helper`
-----------------------------
"""

from __future__ import unicode_literals
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

from zsl import inject, Config, Injected


@inject(config=Config)
def send_email(sender, receivers, subject, text=None, html=None, charset='utf-8', config=Injected):
    """Sends an email.

    :param sender: Sender as string or None for default got from config.
    :param receivers: String or array of recipients.
    :param subject: Subject.
    :param text: Plain text message.
    :param html: Html message.
    :param charset: Charset.
    :param config: Current configuration
    """
    smtp_config = config['SMTP']

    # Receivers must be an array.
    if not isinstance(receivers, list) and not isinstance(receivers, tuple):
        receivers = [receivers]

    # Create the messages
    msgs = []
    if text is not None:
        msgs.append(MIMEText(text, 'plain', charset))

    if html is not None:
        msgs.append(MIMEText(html, 'html', charset))

    if len(msgs) == 0:
        raise Exception("No message is given.")

    if len(msgs) == 1:
        msg = msgs[0]
    else:
        msg = MIMEMultipart()
        for m in msgs:
            msg.attach(m)

    # Default sender.
    if sender is None:
        sender = smtp_config['SENDER']

    # Fill the info.
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = ", ".join(receivers)

    # Send.
    smtp_server = smtplib.SMTP(**(smtp_config['SERVER']))
    smtp_server.sendmail(sender, receivers, msg.as_string())
    smtp_server.quit()



"""
Test the `transactional_guard` decorator in terms of creating the transaction
before checking the policies and calling rollback if policy is broken.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

from mocks import mock_db_session
from zsl.application.modules.alchemy_module import TransactionHolder

try:
    import unittest.mock as mock
except ImportError:
    import mock

from unittest import TestCase

from zsl import Zsl
from zsl.application.containers.web_container import WebContainer
from zsl.resource.model_resource import ModelResource
from zsl.resource.guard import transactional_guard, GuardedMixin, \
    ResourcePolicy, Access

from resource.resource_test_helper import UserModel, \
    create_resource_test_data, users
from zsl.testing.db import IN_MEMORY_DB_SETTINGS, DbTestCase


class UserResource(ModelResource):
    __model__ = UserModel


class TransactionalGuardTest(TestCase):
    def setUp(self):
        zsl = Zsl(__name__, config_object=IN_MEMORY_DB_SETTINGS,
                  modules=WebContainer.modules())
        zsl.testing = True

        create_resource_test_data()
        super(TransactionalGuardTest, self).setUp()

    def testIsInTransaction(self):
        test_case = self

        class AllowPolicy(ResourcePolicy):
            default = Access.ALLOW

        @transactional_guard([AllowPolicy()])
        class GuardedUserModel(UserResource, GuardedMixin):
            def secure_read(self, *args, **kwargs):
                test_case.assertIsNotNone(self._orm)
                test_case.assertTrue(self._in_transaction)

                return super(GuardedUserModel, self).read(*args, **kwargs)

        resource = GuardedUserModel()
        user = resource.read('1', {}, {})

        self.assertDictEqual(users[0]._asdict(), user.get_attributes(),
                             "should return firs user")

    @staticmethod
    def testRollbackBefore():
        class DenyPolicy(ResourcePolicy):
            default = Access.DENY

        @transactional_guard([DenyPolicy()])
        class GuardedUserModel(UserResource, GuardedMixin):
            pass

        class TestTHolder(TransactionHolder):
            rollback = mock.MagicMock()
            _orm = mock.MagicMock()

        with mock.patch(
            'zsl.application.modules.alchemy_module.TransactionHolder',
            side_effect=TestTHolder
        ):
            resource = GuardedUserModel()
            resource.read('', {}, {})

        TestTHolder.rollback.assert_called()
        TestTHolder._orm.assert_not_called()

    @staticmethod
    def testRollbackAfter():
        class DenyAfterPolicy(ResourcePolicy):
            default = Access.ALLOW

            def can_read__after(self, *args, **kwargs):
                return Access.DENY

        mock_sess = mock_db_session()

        @transactional_guard([DenyAfterPolicy()])
        class GuardedUserModel(UserResource, GuardedMixin):
            pass

        class MyTestCase(DbTestCase, TestCase):
            def runTest(self):
                pass

            def testIt(self):
                resource = GuardedUserModel()
                resource.read('', {}, {})

                mock_sess.query.assert_called()
                mock_sess.rollback.assert_called()

        test_case = MyTestCase()
        test_case.setUp()
        test_case.testIt()
        test_case.tearDown()



"""
:mod:`zsl.router.method`
------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from builtins import *

import logging

from flask.wrappers import Response
from flask import request

from zsl.application.modules.web.configuration import MethodConfiguration
from zsl.constants import MimeType
from zsl.interface.webservice.utils.execution import execute_web_task
from zsl.interface.webservice.utils.request_data import extract_data
from zsl.interface.webservice.utils.response_headers import append_headers
from zsl.db.model.app_model_json_encoder import AppModelJSONEncoder
import json

from zsl import inject, Config, Zsl, Injected
from zsl.task.job_context import WebJobContext

METHOD_CONFIG_NAME = 'METHOD'


@append_headers
def default_web_responder(rv):
    if isinstance(rv, Response):
        return rv
    return Response(json.dumps(rv, cls=AppModelJSONEncoder),
                    mimetype=MimeType.APPLICATION_JSON.value)


def identity_responder(rv):
    return rv


def set_default_responder(responder):
    global _default_responder_method
    _default_responder_method = responder


_default_responder_method = default_web_responder


class Performer(object):
    def __init__(self, f):
        global _default_responder_method
        self._f = f
        self.__name__ = "zsl-method-router-performer-of-" + f.__name__
        self.__doc__ = f.__doc__ if hasattr(f, '__doc__') else None
        self._responder = None
        self.set_responder(_default_responder_method)

    def set_responder(self, responder):
        self._responder = responder

    def _call_inner_function(self, a, kw):
        return self._f(*a, **kw)

    def __call__(self, *a, **kw):
        jc = WebJobContext(None, extract_data(request), None, None, request)
        return execute_web_task(jc, lambda: self._call_inner_function(a, kw))


def _get_method_configuration(config):
    # type: (Config) -> MethodConfiguration
    return config.get(METHOD_CONFIG_NAME, MethodConfiguration())


@inject(app=Zsl, config=Config)
def route(path, app=Injected, config=Injected, **options):
    def _decorator(f):
        method_config = _get_method_configuration(config)
        url = "/{0}{1}".format(method_config.url_prefix, path)
        logging.getLogger(__name__).info(
            "Mapping url '{0}' as a method.".format(url))
        routed_function = app.route(url, **options)
        return routed_function(Performer(f))

    return _decorator


@inject(config=Config)
def get_method_packages(config):
    return _get_method_configuration(config).packages



from zsl.utils.xml_to_json import xml_to_json
import unittest

import xml.etree.cElementTree


class TestXmlToJson(unittest.TestCase):
    xml = '''<?xml version="1.0"?>
<root atr="34">
    <simple id="hou"/>
    <node_a>
        text
    </node_a>
    
    <complex>
        <node_b>
            krajina
        </node_b>
        
        <node_c hud="klud"/>
    </complex>
    
    <node_d id="23"/>
    
    <list>
        <item prop="1">34</item>
        <item prop="0">35</item>
        <item prop="1">38</item>
        <item prop="0">39</item>
    </list>
</root>
    '''

    def setUp(self):
        self._root = xml.etree.cElementTree.fromstring(self.xml)

    def test_simple_definitions(self):
        def_attribute = '@atr'
        def_node = 'node_a'
        def_complex_node = 'complex/node_b'

        self.assertEqual("34", xml_to_json(self._root, def_attribute))
        self.assertEqual("text", xml_to_json(self._root, def_node))
        self.assertEqual("krajina", xml_to_json(self._root, def_complex_node))

    def test_func_callback(self):
        def func(element):
            return element.tag

        self.assertEquals('root', xml_to_json(self._root, func))

    def test_tuple_definitions(self):
        self.assertEqual('23', xml_to_json(self._root, ('node_d', '@id')))
        self.assertEqual('klud', xml_to_json(self._root, ('complex/node_c', '@hud')))
        self.assertEqual('textkrajina', xml_to_json(self._root, (lambda x, y: x + y, 'node_a', 'complex/node_b')))

    def test_list_definitions(self):
        self.assertListEqual(['34', '35', '38', '39'], xml_to_json(self._root, ['list/item']))
        self.assertListEqual(['1', '0', '1', '0'], xml_to_json(self._root, ['list/item', '@prop']))

    def test_dict_definitions(self):
        definition = {
            'id': '@atr',
            'a': 'node_a',
            'b': ('node_d', '@id'),
            'c': 'complex/node_b'
        }

        expected = {'id': '34', 'a': 'text', 'b': '23', 'c': 'krajina'}

        self.assertDictEqual(expected, xml_to_json(self._root, definition))


if __name__ == "main":
    unittest.main()



"""
:mod:`zsl.interface.webservice.performers.default`
--------------------------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals

import http.client
import logging

from zsl import inject, Zsl


@inject(app=Zsl)
def create_not_found_mapping(app):
    @app.route("/", defaults={'path': ''})
    @app.route("/<path:path>")
    def not_found_mapping(path):
        """
        Default web request handler. Only returns 404 status code.
        """

        response_str = "404 Not Found: path '{0}' was not mapped.".format(path)
        logging.warning(response_str)
        return response_str, http.client.NOT_FOUND



"""
:mod:`zsl.interface.webservice.performers.resource`
---------------------------------------------------
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from future.utils import viewitems

import json
import logging

from flask import request, Response

from zsl import Zsl, inject
from zsl.application.error_handler import error_handler
from zsl.constants import MimeType
from zsl.db.model import AppModelJSONEncoder
from zsl.interface.webservice.utils.execution import notify_responders, \
    convert_to_web_response
from zsl.interface.webservice.utils.response_headers import append_headers
from zsl.utils.request_helper import args_to_dict
from zsl.utils.resource_helper import parse_resource_path, get_resource_task
from zsl.interface.resource import ResourceResult


@inject(app=Zsl)
def create_resource_mapping(app):
    @app.route("/resource/<path:path>",
               methods=['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS'])
    @append_headers
    @convert_to_web_response
    @error_handler
    def perform_resource(path):
        logging.debug("Getting resource %s.", path)

        if request.method == 'OPTIONS':
            response = app.make_default_options_response()
        else:
            (resource, params) = parse_resource_path(path)
            resource_task = get_resource_task(resource)
            if resource_task is None:
                raise ImportError("No resource named '{0}'.".format(resource))
            msg_format = "Fetched resource named '{0}' with data\n{1}."
            logging.debug(msg_format.format(resource, request.data))

            data = request.get_json() if request.data else None

            resource_result = resource_task(params=params,
                                            args=args_to_dict(request.args),
                                            data=data)

            if not isinstance(resource_result, ResourceResult):
                resource_result = ResourceResult(body=resource_result)

            response = Response(
                json.dumps(resource_result.body, cls=AppModelJSONEncoder),
                mimetype=MimeType.APPLICATION_JSON.value)

            if resource_result.status:
                response.status = str(resource_result.status)

            if resource_result.location:
                response.location = resource_result.location

            if resource_result.count is not None:
                response.headers['X-Total-Count'] = resource_result.count

            if resource_result.links:
                response.headers['Links'] = ', '.join(
                    ['<{url}>; rel="{name}"'.format(url=url, name=name)
                     for name, url in viewitems(resource_result.links)]
                )

        return response



"""
:mod:`zsl.tasks.asl.cors_test_task`
-----------------------------------

Created on 22.12.2012

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals

from builtins import object
from zsl.task.task_decorator import json_input, json_output, crossdomain
from zsl import Zsl
from injector import inject


class CorsTestTask(object):

    @inject(app=Zsl)
    def __init__(self, app):
        self._app = app

    @json_input
    @json_output
    @crossdomain(methods=["GET", "OPTIONS"])
    def perform(self, data):
        return "ok"



"""
:mod:`zsl.resource.model_resource` -- REST for a DB model.
----------------------------------------------------------

Resources provide a way how to directly access DB tables (raw models) and
perform CRUD operations upon them. The default classes of the models should be
overridden to provide more logic or restrictions if wanted.

The basic way to use them is as follows:
 - The models can be defined in the __exposer__.py which is set up at settings.
 - To override just override the basic methods - `create`, `read`, `update`,
   `delete`.

.. moduleauthor:: Peter Morihladko <peter@atteq.com>, Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals

from builtins import object
from builtins import int
from future.utils import viewitems

import logging
import json
from hashlib import sha256
from typing import Union, List, Any

from sqlalchemy.orm import class_mapper

from zsl.resource.resource_helper import filter_from_url_arg, apply_related, create_related_tree, \
    related_from_fields, order_from_url_arg
from zsl.db.helpers import app_models
from zsl.db.helpers.nested import nested_models, nested_model
from zsl.service.service import transactional, TransactionalSupportMixin
from zsl import inject
from zsl.cache.cache_module import CacheModule
from zsl.cache.id_helper import IdHelper, create_key_class_prefix
from zsl.utils.cache_helper import app_model_encoder_fn, app_model_decoder_fn
from zsl.db.model.app_model import AppModel


def dict_pick(dictionary, allowed_keys):
    """
    Return a dictionary only with keys found in `allowed_keys`
    """
    return {key: value for key, value in viewitems(dictionary) if key in allowed_keys}


def page_to_offset(params):
    """
    Transforms `page`/`per_page` from `params` to `limit`/`offset` suitable for SQL.

    :param dict params: The dictionary containing `page` and `per_page` values will be added
                        the values `limit` and `offset`.
    """

    if 'page' not in params:
        return

    page = params['page']
    del params['page']

    # 'per_page' je len alias za 'limit'
    if 'per_page' in params:
        per_page = params.get('per_page')
        del params['per_page']
    else:
        per_page = params.get('limit', 10)

    params['offset'] = int(per_page) * (int(page) - 1)
    params['limit'] = per_page


def _is_list(list_):
    # type: (Any) -> bool
    """Test if variable is a list.

    Maybe this is not the best way to test for a list, but it is sufficient for
    current use case.
    """
    try:
        len(list_)
        return True
    except TypeError:
        return False


class ResourceQueryContext(object):
    """
    The context of the resource query.
     - holds the parameters and arguments of the query,
     - holds the related models which should be fetched (parsed from the arguments),
     - holds the given filter and splits it to the given field array (parsed from the arguments)

    .. automethod:: __init__
    """

    def __init__(self, params, args, data):
        # type: (dict, list, dict) -> ()
        self._args = args.copy()
        self._args_original = args
        self._data = data

        self._params = params if _is_list(params) else [params]

        # Prepare fields and related.
        if 'related' in self._args:
            self._args['related'] = self._args['related'].split(',')

        if 'fields' in self._args:
            self._args['fields'] = self._args['fields'].split(',')
            # we can pass related fields with this, ensure its in 'related' union of two lists
            self._args['related'] = list(
                set(self._args.get('related', [])) | set(related_from_fields(self._args['fields'])))

    @property
    def params(self):
        """Params are given as the part of the path in URL. For example GET /entities/1 will have.
        1 in the params.
        """
        return self._params

    @property
    def args(self):
        """Args are in the query part of the url ?related=&filter_by etc."""
        return self._args

    @property
    def data(self):
        """Body of the request."""
        return self._data

    def get_row_id(self):
        """First parameter, if given, else None. Handy for GET requests."""
        return None if len(self.params) == 0 else self.params[0]

    def get_related(self):
        """Related argument - parsed as array, original argument containing the
        list of comma delimited models which should be fetched along with the resource.
        """
        return self._args.get('related', None)

    def get_filter_by(self):
        """Filter argument - list of filters."""
        return self._args.get('filter_by', None)


class ModelResourceBase(TransactionalSupportMixin):
    """ModelResource works only for tables with a single-column identifier
    (key).

    .. automethod:: __init__
    """

    def __init__(self, model_cls=None):
        """
        Create Model CRUD resource for ``model_cls``
        """

        super(ModelResourceBase, self).__init__()

        if not model_cls:
            self.model_cls = self.__model__
        else:
            self.model_cls = model_cls

        mapper = class_mapper(self.model_cls)
        self._model_pk = mapper.primary_key[0]
        self._model_columns = [column.key for column in mapper.column_attrs]
        self._related_columns = [column.key for column in mapper.relationships]

    def to_filter(self, query, arg):
        return filter_from_url_arg(self.model_cls, query, arg)

    def add_related(self, query, related):
        return apply_related(self.model_cls, query, related)

    def set_ordering(self, query, arg):
        return order_from_url_arg(self.model_cls,  query, arg)


class ModelResource(ModelResourceBase):
    """

    .. automethod:: _create_context
    .. automethod:: _create_one
    .. automethod:: _save_one
    .. automethod:: _delete_one
    .. automethod:: _create_delete_one_query
    """
    @staticmethod
    def _create_context(params, args, data):
        # type: (dict, list, dict) -> ResourceQueryContext
        """
        Creates the resource query context - this an object holding the data
        alongside the querying of the resource. This object is always present as
        a parameter for each method during the query and users.py are free to
        create own properties so that they can optimize and perform the query
        (so the subsequent methods have an access to the already precomputed
        data).
        """
        return ResourceQueryContext(params, args, data)

    @transactional
    def create(self, params, args, data):
        # type: (str, dict, dict) -> AppModel
        """
        POST /resource/model_cls/
        data

        Create new resource
        """
        ctx = self._create_context(params, args, data)
        model = self._create_one(ctx)
        self._save_one(model, ctx)
        return self._return_saved_one(model, ctx)

    def read(self, params=None, args=None, data=None):
        # type: (str, dict, dict) -> Union[List[AppModel], AppModel]
        """
        GET /resource/model_cls/[params:id]?[args:{limit,offset,page,per_page,filter_by,order_by,related,fields}]

        Get resource/s

        :param params
        :type params list
        :param args
        :type args dict
        :param data
        :type data: dict
        """
        if params is None:
            params = []
        if args is None:
            args = {}

        ctx = self._create_context(params, args, data)
        row_id = ctx.get_row_id()

        if row_id:
            return self._get_one(row_id, ctx)

        elif 'count' in args:
            return self._get_collection_count(ctx)

        elif 'desc' in args:
            return self._get_collection_desc(ctx)

        else:
            if 'page' in ctx.args:
                page_to_offset(ctx.args)

            return self._get_collection(ctx)

    @transactional
    def update(self, params, args, data):
        # type: (str, dict, dict) -> Union[List[AppModel], AppModel]
        """
        PUT /resource/model_cls/[params:id]
        data

        Update resource/s
        """
        ctx = self._create_context(params, args, data)
        row_id = ctx.get_row_id()

        if row_id is not None:
            model = self._update_one(ctx)
            return None if model is None else model.get_app_model()
        else:
            return app_models(self._update_collection(ctx))

    @transactional
    def delete(self, params, args, data):
        # type: (str, dict, dict) -> None
        """
        DELETE /resource/model_cls/[params]?[args]

        delete resource/s
        """
        ctx = self._create_context(params, args, data)
        row_id = ctx.get_row_id()

        if row_id is not None:
            return self._delete_one(row_id, ctx)
        else:
            return self._delete_collection(ctx)

    # Create implementation
    def _create_one(self, ctx):
        """
        Creates an instance to be saved when a model is created.
        """
        assert isinstance(ctx, ResourceQueryContext)

        fields = dict_pick(ctx.data, self._model_columns)
        model = self.model_cls(**fields)
        return model

    def _save_one(self, model, ctx):
        """
        Saves the created instance.
        """
        assert isinstance(ctx, ResourceQueryContext)

        self._orm.add(model)
        self._orm.flush()

    @staticmethod
    def _return_saved_one(model, ctx):
        """
        Returns the result of the create operation.
        """
        return model.get_app_model()

    # Read one implementation.
    @transactional
    def _get_one(self, row_id, ctx):
        assert isinstance(ctx, ResourceQueryContext)

        q = self._orm.query(self.model_cls).filter(self._model_pk == row_id)

        related = ctx.get_related()
        if related is not None:
            q = self.add_related(q, related)
            return nested_model(self._read_one(q, ctx),
                                create_related_tree(related))
        else:
            return self._read_one(q, ctx).get_app_model()

    @staticmethod
    def _read_one(q, ctx):
        return q.one()

    # Read collection implementation.
    def _create_collection_query(self, ctx):
        return self._orm.query(self.model_cls)

    @transactional
    def _get_collection_count(self, ctx):
        assert isinstance(ctx, ResourceQueryContext)

        filter_by = ctx.get_filter_by()

        q = self._create_collection_query(ctx)

        if filter_by is not None:
            q = self.to_filter(q, filter_by)

        return self._read_collection_count(q, ctx)

    def _read_collection_count(self, q, ctx):
        return q.with_entities(self._model_pk).count()

    @transactional
    def _get_collection(self, ctx):
        assert isinstance(ctx, ResourceQueryContext)

        order_by = ctx.args.get('order_by', None)
        filter_by = ctx.get_filter_by()
        related = ctx.get_related()

        q = self._create_collection_query(ctx)

        if filter_by is not None:
            q = self.to_filter(q, filter_by)

        if order_by is not None:
            q = self.set_ordering(q, order_by)

        if related is not None:
            q = self.add_related(q, related)
            return nested_models(self._read_collection(q, ctx),
                                 create_related_tree(related))
        else:
            return app_models(self._read_collection(q, ctx))

    @staticmethod
    def _read_collection(q, ctx):
        offset = int(ctx.args.get('offset', 0))
        limit = ctx.args.get('limit', 10)
        if limit is not None and limit != 'unlimited':
            q = q.limit(int(limit))
        if offset > 0:
            q = q.offset(offset)
        return q.all()

    def _get_collection_desc(self):
        return [column.name for column in class_mapper(self.model_cls).columns]

    # Update
    def _update_one_simple(self, row_id, fields, ctx):
        fields = dict_pick(fields, self._model_columns)

        model = self._orm.query(self.model_cls).get(row_id)

        if model is None:
            return None

        for name, value in viewitems(fields):
            setattr(model, name, value)

        return model

    def _update_one(self, ctx):
        """
        Update row
        """
        assert isinstance(ctx, ResourceQueryContext)
        fields = ctx.data
        row_id = ctx.get_row_id()
        return self._update_one_simple(row_id, fields, ctx)

    def _update_collection(self, ctx):
        """
        Bulk update
        """
        assert isinstance(ctx, ResourceQueryContext)
        models = []

        for row in ctx.data:
            models.append(self._update_one_simple(row.pop('id'), row, ctx))

        return models

    # Delete methods
    def _delete_one(self, row_id, ctx):
        """
        Deletes row by the given id -- `row_id`. The method first created the
        query using the method :meth:`_create_delete_one_query` and then executes it.

        :param int row_id: Identifier of the deleted row.
        :param ResourceQueryContext ctx: The context of this delete query.
        """
        return self._create_delete_one_query(row_id, ctx).delete()

    def _create_delete_one_query(self, row_id, ctx):
        """
        Delete row by id query creation.

        :param int row_id: Identifier of the deleted row.
        :param ResourceQueryContext ctx: The context of this delete query.
        """

        assert isinstance(ctx, ResourceQueryContext)
        return self._orm.query(self.model_cls).filter(self._model_pk == row_id)

    def _delete_collection(self, ctx):
        """
        Delete a collection from DB, optionally filtered by ``filter_by``
        """
        assert isinstance(ctx, ResourceQueryContext)

        filter_by = ctx.get_filter_by()
        q = self._orm.query(self.model_cls)

        if filter_by is not None:
            q = self.to_filter(q, filter_by)

        return q.delete()


class CachedModelResource(ModelResource):
    """
    The cached resource - uses redis to cache the resource for the given amount of seconds.
    """

    @inject(cache_module=CacheModule, id_helper=IdHelper, logger=logging.Logger)
    def __init__(self, model_cls, cache_module, id_helper, logger, timeout='short'):
        super(CachedModelResource, self).__init__(model_cls)
        self._cache_module = cache_module
        self._id_helper = id_helper
        self._logger = logger
        self._timeout = timeout

    def _create_key(self, arghash):
        key_prefix = create_key_class_prefix(self.model_cls)
        return "cached-resource:{0}:{1}".format(key_prefix, arghash)

    def _create_key_from_context(self, ctx):
        arghash = sha256(json.dumps({'params': ctx.params, 'args': ctx.args, 'data': ctx.data})).hexdigest()
        return self._create_key(arghash)

    def _get_one(self, row_id, ctx):
        # Make hash of params, args and data and ache using the hash in the key.
        key = self._create_key_from_context(ctx)
        self._logger.debug("CachedModelResource - get one, key: {0}.".format(key))

        if self._id_helper.check_key(key):
            result = json.loads(self._id_helper.get_key(key))
        else:
            self._logger.debug("CachedModelResource - get one not cached, transferring to resource...")
            result = super(CachedModelResource, self)._get_one(row_id, ctx)
            # serialize as model
            self._id_helper.set_key(key, app_model_encoder_fn(result), self._timeout)

        self.invalidate()

        return result

    def _get_collection_count(self, ctx):
        # Make hash of params, args and data and ache using the hash in the key.
        key = self._create_key_from_context(ctx)
        self._logger.debug("CachedModelResource - get one, key: {0}.".format(key))

        if self._id_helper.check_key(key):
            result = int(self._id_helper.get_key(key))
        else:
            self._logger.debug("CachedModelResource - get one not cached, transferring to resource...")
            result = super(CachedModelResource, self)._get_collection_count(ctx)
            self._id_helper.set_key(key, app_model_encoder_fn(result), self._timeout)

        return result

    def _get_collection(self, ctx):
        # Make hash of params, args and data and ache using the hash in the key.
        key = self._create_key_from_context(ctx)
        self._logger.debug("CachedModelResource - collection, key: {0}.".format(key))

        if self._id_helper.check_key(key):
            result = self._id_helper.gather_page(key, app_model_decoder_fn)
        else:
            self._logger.debug("CachedModelResource - collection not cached, transferring to resource...")
            result = super(CachedModelResource, self)._get_collection(ctx)
            self._id_helper.fill_page(key, result, self._timeout, app_model_encoder_fn)

        return result

    def invalidate(self):
        """
        Invalidates all the data associated with this model
        """
        key = self._create_key("")
        self._id_helper.invalidate_keys_by_prefix(key)

    def create(self, params, args, data):
        rv = ModelResource.create(self, params, args, data)
        self.invalidate()
        return rv

    def update(self, params, args, data):
        rv = ModelResource.update(self, params, args, data)
        self.invalidate()
        return rv

    def delete(self, params, args, data):
        rv = ModelResource.delete(self, params, args, data)
        self.invalidate()
        return rv


class ReadOnlyResourceUpdateOperationException(Exception):
    def __init__(self, operation):
        self._operation = operation
        super(ReadOnlyResourceUpdateOperationException, self).__init__(
            "Can not perform operation '{0}' on ReadOnlyResource.".format(operation))

    def get_operation(self):
        return self._operation

    operation = property(get_operation)


class ReadOnlyResourceMixin(object):
    """
    The mixin to be used to forbid the update/delete and create operations.
    Remember the Python's MRO and place this mixin at the right place in the inheritance declaration.

    .. automethod:: create
    .. automethod:: update
    .. automethod:: delete 
    """

    OPERATION_CREATE = 'create'
    OPERATION_UPDATE = 'update'
    OPERATION_DELETE = 'delete'

    @staticmethod
    def create(params, args, data):
        """Raises exception.
        
        Just raises ReadOnlyResourceUpdateOperationException to indicate
        that this method is not available.

        :raises ReadOnlyResourceUpdateOperationException: when accessed  
        """
        raise ReadOnlyResourceUpdateOperationException(ReadOnlyResourceMixin.OPERATION_CREATE)

    @staticmethod
    def update(params, args, data):
        """Raises exception.
        
        Just raises ReadOnlyResourceUpdateOperationException to indicate
        that this method is not available.

        :raises ReadOnlyResourceUpdateOperationException: when accessed  
        """
        raise ReadOnlyResourceUpdateOperationException(ReadOnlyResourceMixin.OPERATION_UPDATE)

    @staticmethod
    def delete(params, args, data):
        """Raises exception.
        
        Just raises ReadOnlyResourceUpdateOperationException to indicate
        that this method is not available.

        :raises ReadOnlyResourceUpdateOperationException: when accessed  
        """
        raise ReadOnlyResourceUpdateOperationException(ReadOnlyResourceMixin.OPERATION_DELETE)



"""
:mod:`zsl.db.helpers.pagination`
--------------------------------
"""
from __future__ import unicode_literals
from builtins import object

from sqlalchemy.orm.query import Query

FIRST_PAGE = 1
DEFAULT_PAGE_RECORD_COUNT = 25


class Pagination(object):
    def __init__(self, pagination=None):
        pagination = self._create_pagination_model(pagination)
        assert isinstance(pagination, PaginationModel)
        self._offset = (pagination.page_no - FIRST_PAGE) * pagination.page_record_count
        self._page_record_count = pagination.page_record_count
        self._record_count = 0

    def _create_pagination_model(self, pagination):
        if pagination is None:
            pagination = PaginationModel()
        elif isinstance(pagination, dict):
            page_record_count = int(pagination.get('page_record_count', DEFAULT_PAGE_RECORD_COUNT))
            pagination = PaginationModel(FIRST_PAGE + int(pagination.get('offset', 0)) // page_record_count,page_record_count)
        return pagination

    def set_record_count(self, record_count):
        self._record_count = record_count

    def get_record_count(self):
        return self._record_count

    def get_page_record_count(self):
        return self._page_record_count

    def get_offset(self):
        count = self.get_record_count()
        per_page = self.get_page_record_count()
        if self._offset >= count:
            last_page_size = count % per_page
            if last_page_size == 0:
                last_page_size = per_page
            self._offset = count - last_page_size
        if self._offset < 0:
            self._offset = 0

        return self._offset

    def apply_pagination(self, q):
        # type: (Query)->Query
        return q.limit(self.get_page_record_count()).offset(self.get_offset())

    def paginate(self, q):
        self.set_record_count(q.count())
        return self.apply_pagination(q).all()


class PaginationModel(object):
    def __init__(self, page_no=FIRST_PAGE,
                 page_record_count=DEFAULT_PAGE_RECORD_COUNT):
        self.page_no = int(page_no if page_no else FIRST_PAGE)
        self.page_record_count = int(
            page_record_count if page_record_count else DEFAULT_PAGE_RECORD_COUNT)



"""
:mod:`zsl.utils.rss`
--------------------

Helper function for handling rss.

.. moduleauthor:: Peter Morihladko
"""
from __future__ import unicode_literals
from future.utils import viewitems

import xml.etree.cElementTree as ET


def complex_el_from_dict(parent, data, key):
    """Create element from a dict definition and add it to ``parent``.

    :param parent: parent element
    :type parent: Element
    :param data: dictionary with elements definitions, it can be a simple \
    {element_name: 'element_value'} or complex \
    {element_name: {_attr: {name: value, name1: value1}, _text: 'text'}}
    :param key: element name and key in ``data``
    :return: created element
    """
    el = ET.SubElement(parent, key)
    value = data[key]

    if isinstance(value, dict):
        if '_attr' in value:
            for a_name, a_value in viewitems(value['_attr']):
                el.set(a_name, a_value)

        if '_text' in value:
            el.text = value['_text']

    else:
        el.text = value

    return el


def element_from_dict(parent, data, element):
    """Create ``element`` to ``parent`` and sets its value to data[element], which
    will be removed from the ``data``.

    :param parent: parent element
    :type parent: Element
    :param data: dictionary where data[element] is desired value
    :type data: dict(str, str)
    :param element: name of the new element
    :type element: str
    :return: created element
    """
    el = ET.SubElement(parent, element)
    el.text = data.pop(element)

    return el


def rss_create(channel, articles):
    """Create RSS xml feed.

    :param channel: channel info [title, link, description, language]
    :type channel: dict(str, str)
    :param articles: list of articles, an article is a dictionary with some \
    required fields [title, description, link] and any optional, which will \
    result to `<dict_key>dict_value</dict_key>`
    :type articles: list(dict(str,str))
    :return: root element
    :rtype: ElementTree.Element
    """
    channel = channel.copy()

    # TODO use deepcopy
    # list will not clone the dictionaries in the list and `elemen_from_dict`
    # pops items from them
    articles = list(articles)

    rss = ET.Element('rss')
    rss.set('version', '2.0')

    channel_node = ET.SubElement(rss, 'channel')

    element_from_dict(channel_node, channel, 'title')
    element_from_dict(channel_node, channel, 'link')
    element_from_dict(channel_node, channel, 'description')
    element_from_dict(channel_node, channel, 'language')

    for article in articles:
        item = ET.SubElement(channel_node, 'item')

        element_from_dict(item, article, 'title')
        element_from_dict(item, article, 'description')
        element_from_dict(item, article, 'link')

        for key in article:
            complex_el_from_dict(item, article, key)

    return ET.ElementTree(rss)



"""
:mod:`zsl.cache.redis_cache_module`
-----------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from zsl.cache.cache_module import CacheModule
import redis

from zsl import inject, Zsl, Config


class RedisCacheModule(CacheModule):
    """Abstraction layer for caching."""

    @inject(app=Zsl, config=Config)
    def __init__(self, app, config):
        """Abstraction layer for caching."""
        self._app = app
        self._config = config

        redis_conf = self._config.get('REDIS', {})

        self._client = redis.StrictRedis(
            host=redis_conf.get('host'),
            port=redis_conf.get('port'),
            db=redis_conf.get('db', 0),
            password=redis_conf.get('password')
        )
        self.logger = self._app.logger.getChild('cache')
        self.logger.debug("Redis client created.")
        self._cache_prefix = self._config[
            'CACHE_PREFIX'] + ':' if 'CACHE_PREFIX' in self._config else ''

    def _prefix_key(self, key):
        return self._cache_prefix + key

    def set_key(self, key, value, timeout):
        pkey = self._prefix_key(key)
        self._client.set(pkey, value)
        self.set_key_expiration(key, timeout)

    def invalidate_key(self, key):
        pkey = self._prefix_key(key)
        self.logger.debug("Key invalidation '{0}'.".format(key))
        self._client.delete(pkey)

    def set_key_expiration(self, key, timeout):
        pkey = self._prefix_key(key)
        self.logger.debug("Key expiration '{0}' = {1}.".format(key, timeout))
        self._client.expire(pkey, timeout)

    def contains_key(self, key):
        pkey = self._prefix_key(key)
        return self._client.exists(pkey)

    def contains_list(self, key):
        pkey = self._prefix_key(key)
        return self._client.exists(pkey)

    def get_key(self, key):
        pkey = self._prefix_key(key)
        return self._client.get(pkey)

    def append_to_list(self, key, value):
        pkey = self._prefix_key(key)
        self._client.rpush(pkey, value)

    def get_list(self, key):
        pkey = self._prefix_key(key)
        llen = self._client.llen(pkey)
        return self._client.lrange(pkey, 0, llen - 1)

    def invalidate_by_glob(self, glob):
        pglob = self._prefix_key(glob)

        if self._config.get('IS_USING_MEDIEVAL_SOFTWARE', False):
            keylist = self._client.keys(pglob)
        else:
            keylist = self._client.scan_iter(pglob)

        for key in keylist:
            # This does not need to prefixed.
            self.logger.debug('Invalidating key {0}.'.format(key))
            self._client.delete(key)



"""
:mod:`zsl.resource.guard`
-------------------------

Guard module defines tools to inject security checks into a resource. With
help of the ``guard`` class decorator and ``ResourcePolicy`` declarative
policy class a complex security resource behaviour can be achieved.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from future.utils import raise_from
from builtins import *

import http.client

from typing import List, Optional, Dict, Any, Callable
from functools import wraps

from zsl.interface.resource import ResourceResult
from zsl.service.service import SessionFactory, transactional, \
    _TX_HOLDER_ATTRIBUTE

from enum import Enum

from zsl.utils.http import get_http_status_code_value

_HTTP_STATUS_FORBIDDEN = get_http_status_code_value(http.client.FORBIDDEN)


class Access(Enum):
    ALLOW = 1
    DENY = 2
    CONTINUE = 3


class ResourcePolicy(object):
    """Declarative policy class.

    Every CRUD method has is corespondent *can_method__before* and
    *can_method__after* where *method* can be one of (*create*, *read*,
    *update*, *delete*). *__before* method will get the CRUD method
    parameters and *__after* will get the CRUD method result as parameter. On
    returning ``Access.ALLOW`` access is granted. It should return
    ``Access.CONTINUE`` when the policy is not met, but is not broken, i. e. it
    is not its responsibility to decide. On returning ``Access.DENY`` or raising
    a ``PolicyException`` policy is broken  and access is immediately denied.

    The default implementation of these method lookup for corresponding
    attribute *can_method*, so ``can_read = Access.ALLOW`` will allow access
    for reading without the declaration of ``can_read__before`` or
    ``can_read__after``. *default* attribute is used if *can_method*
    attribute is not declared. For more complex logic it can be declared as a
    property, see examples:

    .. code-block:: python


        class SimplePolicy(ResourcePolicy):
            '''Allow read and create'''

            default = Access.ALLOW
            can_delete = Access.CONTINUE
            can_update = Access.CONTINUE


        class AdminPolicy(ResourcePolicy):
            '''Only admin has access'''

            @inject(user_service=UserService)
            def __init__(self, user_service):
                self._user_service = user_service

            @property
            def default(self):
                if self._user_service.current_user.is_admin:
                    return Access.ALLOW
    """
    default = Access.CONTINUE

    # can_create

    # can_read

    # can_update

    # can_delete

    def can_create__before(self, *args, **kwargs):
        """Check create method before executing."""
        return self._check_default('can_create')

    def can_create__after(self, *args, **kwargs):
        """Check create method after executing."""
        return self._check_default('can_create')

    def can_read__before(self, *args, **kwargs):
        """Check read method before executing."""
        return self._check_default('can_read')

    def can_read__after(self, *args, **kwargs):
        """Check read method after executing."""
        return self._check_default('can_read')

    def can_update__before(self, *args, **kwargs):
        """Check update method before executing."""
        return self._check_default('can_update')

    def can_update__after(self, *args, **kwargs):
        """Check update method after executing."""
        return self._check_default('can_update')

    def can_delete__before(self, *args, **kwargs):
        """Check delete method before executing."""
        return self._check_default('can_delete')

    def can_delete__after(self, *args, **kwargs):
        """Check delete method after executing."""
        return self._check_default('can_delete')

    def _check_default(self, prop):
        # type: (str) -> Access
        return getattr(self, prop, self.default)


class PolicyViolationError(Exception):
    """Error raised when policy is violated.

    It can bear a HTTP status code, 403 is by default.
    """

    def __init__(self, message, code=_HTTP_STATUS_FORBIDDEN):
        self.code = code
        super(PolicyViolationError, self).__init__(message)


class GuardedMixin(object):
    """Add guarded CRUD methods to resource.

    The ``guard`` replaces the CRUD guarded methods with a wrapper with
    security checks around these methods. It adds this mixin into the
    resource automatically, but it can be declared on the resource manually
    for IDEs to accept calls to the guarded methods.
    """

    def guarded_create(self, params, args, data):
        # type: (str, Dict[str, str], Dict[str, Any]) -> Dict[str, Any]
        pass

    def guarded_read(self, params, args, data):
        # type: (str, Dict[str, str], Dict[str, Any]) -> Dict[str, Any]
        pass

    def guarded_update(self, params, args, data):
        # type: (str, Dict[str, str], Dict[str, Any]) -> Dict[str, Any]
        pass

    def guarded_delete(self, params, args, data):
        # type: (str, Dict[str, str], Dict[str, Any]) -> Dict[str, Any]
        pass


def default_error_handler(e, *_):
    # type: (PolicyViolationError, Any) -> ResourceResult
    """Default policy violation error handler.

    It will create an empty resource result with an error HTTP code.
    """

    return ResourceResult(
        status=e.code,
        body={}
    )


class guard(object):
    """Guard decorator.

    This decorator wraps the CRUD methods with security checks before and
    after CRUD method execution, so that the response can be stopped or
    manipulated. The original CRUD methods are renamed to *guarded_method*,
    where *method* can be [*create*, *read*, *update*, *delete*], so by using a
    `GuardedResource` as a base, you can still redeclare the *guarded_methods*
    and won't loose the security checks.

    It takes a list of policies, which will be always checked before and
    after executing the CRUD method.

    Policy is met, when it returns ``Access.ALLOW``, on ``Access.CONTINUE`` it
    will continue to check others and on ``Access.DENY`` or raising a
    ``PolicyViolationError`` access will be restricted. If there is no policy
    which grants the access a ``PolicyViolationError`` is raised and access
    will be restricted.

    Guard can have a custom exception handlers or method wrappers to _wrap the
    CRUD method around.

    .. code-block:: python


        class Policy(ResourcePolicy):
            default = Access.DENY
            can_read = Access.ALLOW  # allow only read


        @guard([Policy()])
        class GuardedResource(GuardedMixin):
            def read(self, param, args, data):
                return resources[param]


        class SpecificResource(GuardedResource):
            # override GuardedResource.read, but with its security checks
            def guarded_read(self, param, args, data):
                return specific_resources[param]

    """
    method_wrappers = []
    exception_handlers = [default_error_handler]
    resource_methods = ['create', 'read', 'update', 'delete']

    def __init__(self, policies=None, method_wrappers=None,
                 exception_handlers=None):
        # type: (Optional[List[policies]]) -> None

        self.policies = list(policies) if policies else []

        if method_wrappers:
            self._method_wrappers = self.method_wrappers + method_wrappers
        else:
            self._method_wrappers = list(self.method_wrappers)

        if exception_handlers:
            self._exception_handlers = self.exception_handlers + \
                                       exception_handlers
        else:
            self._exception_handlers = list(self.exception_handlers)

    @staticmethod
    def _check_before_policies(res, name, *args, **kwargs):
        for policy in res._guard_policies:
            access = _call_before(policy, name)(*args, **kwargs)

            if access == Access.ALLOW:
                return

            elif access == Access.DENY:
                raise PolicyViolationError('Access denied for {} {}'.format(
                    name, 'before'), code=_HTTP_STATUS_FORBIDDEN)

            elif access == Access.CONTINUE:
                continue

            else:
                raise TypeError('Access has no value {}'.format(access))

        raise PolicyViolationError(
            "Access haven't been granted for {} {}".format(
                name, 'before'), code=_HTTP_STATUS_FORBIDDEN)

    @staticmethod
    def _check_after_policies(res, name, result):
        for policy in res._guard_policies:
            access = _call_after(policy, name)(result)

            if access == Access.ALLOW:
                return

            elif access == Access.DENY:
                raise PolicyViolationError('Policy violation for {} {}'.format(
                    name, 'before'), code=_HTTP_STATUS_FORBIDDEN)

            elif access == Access.CONTINUE:
                continue

            else:
                raise TypeError('Access have no value {}'.format(access))

        raise PolicyViolationError(
            "Access haven't been granted for {} {}".format(
                name, 'after'), code=_HTTP_STATUS_FORBIDDEN)

    def _wrap(self, method):
        # type: (Callable) -> Callable

        name = method.__name__

        @wraps(method)
        def wrapped(*args, **kwargs):
            res = args[0]
            args = args[1:]

            try:
                self._check_before_policies(res, name, *args, **kwargs)
                rv = _guarded_method(res, name)(*args, **kwargs)
                self._check_after_policies(res, name, rv)

            except PolicyViolationError as e:
                rv = self._handle_exception(e, res)

            return rv

        for mw in reversed(self._method_wrappers):
            wrapped = mw(wrapped)

        return wrapped

    def _handle_exception(self, error, resource):
        rv = None
        for handler in self._exception_handlers:
            rv = handler(error, rv, resource)

        return rv

    def __call__(self, cls):
        if hasattr(cls, '_guard_policies'):
            self.policies += getattr(cls, '_guard_policies')
            setattr(cls, '_guard_policies', list(self.policies))

            return cls

        setattr(cls, '_guard_policies', list(self.policies))

        for method_name in self.resource_methods:
            guarded_name = _guarded_name(method_name)

            if hasattr(cls, method_name):
                method = getattr(cls, method_name)

                setattr(cls, method_name, self._wrap(method))
                setattr(cls, guarded_name, method)

        if issubclass(cls, GuardedMixin):
            return cls
        else:
            return type(cls.__name__, (cls, GuardedMixin), {})


def transactional_error_handler(e, rv, _):
    # type: (Any, Any, SessionFactory) -> Any
    """Re-raise a violation error to be handled in the
    ``_nested_transactional``.
    """
    raise_from(_TransactionalPolicyViolationError(rv), e)


def _nested_transactional(fn):
    # type: (Callable) -> Callable
    """In a transactional method create a nested transaction."""
    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        # type: (SessionFactory) -> Any

        try:
            rv = fn(self, *args, **kwargs)
        except _TransactionalPolicyViolationError as e:
            getattr(self, _TX_HOLDER_ATTRIBUTE).rollback()
            rv = e.result

        return rv
    return wrapped


class transactional_guard(guard):
    """Security guard for ``ModelResource``.

    This add a transactional method wrapper and error handler which calls the
    rollback on ``PolicyViolationError``.
    """
    method_wrappers = [transactional, _nested_transactional]
    exception_handlers = guard.exception_handlers + [
        transactional_error_handler]


class _TransactionalPolicyViolationError(PolicyViolationError):
    """Exception raised during """
    def __init__(self, result):
        # type: (ResourceResult) -> None
        self.result = result
        super(_TransactionalPolicyViolationError, self).__init__(
            result.body,
            result.status
        )


def _guarded_method(res, method_name):
    # type: (object, str) -> Callable
    """Return the guarded method from CRUD name"""
    return getattr(res, _guarded_name(method_name))


def _guarded_name(method_name):
    # type: (str) -> str
    """Return name for guarded CRUD method.

    >>> _guarded_name('read')
    'guarded_read'
    """
    return 'guarded_' + method_name


def _before_name(method_name):
    # type: (str) -> str
    """Return the name of the before check method.

    >>> _before_name('read')
    'can_read__before'
    """
    return 'can_' + method_name + '__before'


def _after_name(method_name):
    # type: (str) -> str
    """Return the name of after check method.

    >>> _after_name('read')
    'can_read__after'
    """
    return 'can_' + method_name + '__after'


def _call_before(policy, method_name):
    # type: (ResourcePolicy, str) -> Callable
    """Return the before check method.

    >>> p = ResourcePolicy()
    >>> _call_before(p, 'read')
    p.can_read__before
    """
    return getattr(policy, _before_name(method_name))


def _call_after(policy, method_name):
    # type: (ResourcePolicy, str) -> Callable
    """Return the after check method.

    >>> p = ResourcePolicy()
    >>> _call_after(p, 'read')
    p.can_after__before
    """
    return getattr(policy, _after_name(method_name))



# -*- coding: utf-8 -*-
#
# zsl documentation build configuration file, created by
# sphinx-quickstart on Tue Dec 20 16:26:44 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import pkgutil
import sys

src_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, src_path)

from recommonmark.parser import CommonMarkParser

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.intersphinx',
              'sphinx.ext.ifconfig',
              'sphinx.ext.viewcode']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']
# source_suffix = '.rst'
source_parsers = {'.md': CommonMarkParser}

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'zsl'
copyright = u'2016, Atteq s.r.o.'
author = u'Atteq s.r.o.'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = u'0.9a'
# The full version, including alpha/beta/rc tags.
release = u'0.9a'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

import sphinx_rtd_theme

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'zsldoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'zsl.tex', u'zsl Documentation',
     u'Atteq s.r.o.', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'zsl', u'zsl Documentation',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'zsl', u'zsl Documentation',
     author, 'zsl', 'One line description of project.',
     'Miscellaneous'),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}

UNDOC = []

UNDOC_PYTHON_3 = ["zsl.application.containers.gearman_container",
                  "zsl.application.modules.gearman_module",
                  "zsl.interface.gearman.json_data_encoder",
                  "zsl.interface.gearman.task_filler",
                  "zsl.interface.gearman.worker",
                  "zsl.tasks.zsl.cors_test_task",
                  "zsl.tasks.zsl.schedule_gearman_task",
                  "zsl.tasks.zsl.schedule_kill_worker_task",
                  "zsl.utils.background_task",
                  "zsl.utils.gearman_helper"]
UNDOC_PYTHON_2 = []


def api_doc_generate():
    def create_forbidden_modules():
        forbidden_modules = UNDOC
        forbidden_modules += UNDOC_PYTHON_3 if sys.version_info >= (3, 0) else \
            UNDOC_PYTHON_2
        return set(forbidden_modules)

    forbidden_modules = create_forbidden_modules()

    def write_header(f):
        f.writelines(["API\n",
                      "###\n",
                      "\n",
                      ".. toctree::\n",
                      "   :maxdepth: 2\n",
                      "   :caption: Contents:\n",
                      "\n"])

    def is_module_hidden(modname):
        return modname.split('.')[-1].startswith('_')

    def is_module_from_zsl(modname):
        return modname.startswith('zsl.')

    def is_module_forbidden(modname):
        return modname in forbidden_modules

    def gather_modules():

        ignore_error = lambda x: None
        for _, modname, _ in pkgutil.walk_packages(onerror=ignore_error):
            if not is_module_from_zsl(modname):
                continue

            if is_module_hidden(modname):
                continue

            if is_module_forbidden(modname):
                continue

            yield modname

    def write_module(f, modname):
        f.writelines(["\n",
                      ".. automodule:: {0}\n".format(modname),
                      "   :members:\n",
                      "   :undoc-members:\n"])

    def write_modules(f, modnames):
        for modname in modnames:
            write_module(f, modname)

    with open(os.path.join(os.path.dirname(__file__), 'api.rst'), 'w') as f:
        write_header(f)
        modules = list(gather_modules())
        write_modules(f, modules)


api_doc_generate()



"""
:mod:`zsl.utils.command_dispatcher`
-----------------------------------

.. moduleauthor:: peter morihladko
"""
from __future__ import unicode_literals

from builtins import object
import inspect


class CommandDispatcher(object):
    """
    A simple class for command dictionary. A command is a function
    which can take named parameters.
    """

    def __init__(self):
        """
        Create command dictionary
        """

        self.commands = {}

    def command(self, fn):
        """
        Add method or function to dispatcher. Can be use as a nice
        decorator.

        :param fn: function or method
        :type fn: function
        :return: the same function
        :rtype: function
        """
        self.commands[fn.__name__] = fn

        return fn

    """alias for ``CommandDispatcher.command``"""
    add_function = command

    def execute_command(self, command, args=None):
        """
        Execute a command

        :param command: name of the command
        :type command: str
        :param args: optional named arguments for command
        :type args: dict
        :return: the result of command
        :raises KeyError: if command is not found
        """

        if args is None:
            args = {}

        command_fn = self.commands[command]

        return command_fn(**args)

    def bound(self, instance):
        """
        Return a new dispatcher, which will switch all command functions
        with bounded methods of given instance matched by name. It will
        match only regular methods.

        :param instance: object instance
        :type instance: object
        :return: new Dispatcher
        :rtype: CommandDispatcher
        """

        bounded_dispatcher = CommandDispatcher()
        bounded_dispatcher.commands = self.commands.copy()

        for name in self.commands:
            method = getattr(instance, name, None)

            if method and inspect.ismethod(method) and method.__self__ == instance:
                bounded_dispatcher.commands[name] = method

        return bounded_dispatcher



"""
:mod:`zsl.utils.import_helper`
------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
import gearman
import json
from flask.config import Config
from zsl import inject


@inject(config=Config)
def schedule_gearman_task(path, data, config):
    gm_client = gearman.GearmanClient(["{0}:{1}".format(config['GEARMAN']['host'], config['GEARMAN']['port'])])
    gm_client.submit_job(config['GEARMAN_TASK_NAME'], json.dumps(
        {'path': path, 'data': json.dumps(data)}), wait_until_complete=False, background=True)



"""
:mod:`zsl.utils.xml_to_json` -- xml helpers
-------------------------------------------

Helper functions for simpler parsing xml into object with schemas.
"""
from __future__ import unicode_literals
from future.utils import viewitems


class XmlToJsonException(Exception):
    """Exception raised during converting xml to json."""
    # TODO make use of build in exception, see bug #13299
    pass


class NotCompleteXmlException(Exception):
    """Exception raised during parsing an invalid XML."""
    # TODO make use of build in exception, see bug #13299
    pass


def xml_to_json(element, definition, required=False):
    # TODO document tuple - it looks little too complex
    """Convert XML (ElementTree) to dictionary from a definition schema.

    Definition schema can be a simple string - XPath or @attribute for
    direct extraction or a complex one described by

    * dictionary ``{key: 'xpath or @attribute', second: 'complex definition'}`` \
      required parameters can be marked with * at the end
    * list ``[xpath, [definition]]`` - create a list of all elements found by \
      xpath, parse the parts with given definition if provided as second \
      argument
    * Callable - parse the element by given function, can be handy as a part \
      of complex definition

    :param element: ElementTree element
    :type element: ElementTree.Element
    :param definition: schema for the json
    :type definition: Union[str, tuple, dict, list, Callable]
    :param required: parsed value should be not None
    :type required: bool
    :return: parsed xml
    :rtype: Union[dict, str, list]
    """
    # handle simple definition
    if isinstance(definition, str) and len(definition) > 0:
        if definition[0] == '@':  # test for attribute
            return element.get(definition[1:])

        # get tag text
        else:
            sub_element = element.find(definition)

            if sub_element is None:
                if required:
                    raise NotCompleteXmlException('Expecting {0} in element {1}'.format(definition, element.tag))
                return None

            return sub_element.text.strip() if sub_element.text else None

            # handle tuple
    elif isinstance(definition, tuple):
        return _parse_tuple(element, definition, required)

    # handle dict
    elif isinstance(definition, dict):
        return _parse_dict(element, definition)

    # handle list
    elif isinstance(definition, list):
        return _parse_list(element, definition)

    elif hasattr(definition, '__call__'):
        return definition(element)

    # default
    else:
        return element.text.strip() if element.text else None


def _parse_dict(element, definition):
    """Parse xml element by a definition given in dict format.

    :param element: ElementTree element
    :param definition: definition schema
    :type definition: dict
    :return: parsed xml
    :rtype: dict
    """
    sub_dict = {}

    for name, subdef in viewitems(definition):
        (name, required) = _parse_name(name)

        sub_dict[name] = xml_to_json(element, subdef, required)

    return sub_dict


def _parse_tuple(element, definition, required):
    """Parse xml element by a definition given in tuple format.

    :param element: ElementTree element
    :param definition: definition schema
    :type definition: tuple
    :param required: parsed value should be not None
    :type required: bool
    :return: parsed xml
    """
    # TODO needs to be documented properly.
    d_len = len(definition)

    if d_len == 0:
        return None
    if d_len == 1:
        return xml_to_json(element, definition[0], required)

    first = definition[0]

    if hasattr(first, '__call__'):
        # TODO I think it could be done without creating the array
        # first(xml_to_json(element, d) for d in definition[1:]) test it
        return first(*[xml_to_json(element, d) for d in definition[1:]])

    if not isinstance(first, str):
        raise XmlToJsonException('Tuple definition must start with function or string')

    if first[0] == '@':
        raise XmlToJsonException('Tuple definition must not start with attribute')

    sub_elem = element.find(first)

    if sub_elem is None:
        if required:
            raise NotCompleteXmlException('Expecting {0} in element {1}'.format(first, element.tag))

        return None

    return xml_to_json(sub_elem, definition[1], required)


def _parse_list(element, definition):
    """Parse xml element by definition given by list.

    Find all elements matched by the string given as the first value
    in the list (as XPath or @attribute).

    If there is a second argument it will be handled as a definitions
    for the elements matched or the text when not.

    :param element: ElementTree element
    :param definition: definition schema
    :type definition: list
    :return: parsed xml
    :rtype: list
    """
    if len(definition) == 0:
        raise XmlToJsonException('List definition needs some definition')

    tag = definition[0]
    tag_def = definition[1] if len(definition) > 1 else None

    sub_list = []

    for el in element.findall(tag):
        sub_list.append(xml_to_json(el, tag_def))

    return sub_list


def _parse_name(name):
    """Parse name in complex dict definition.

    In complex definition required params can be marked with `*`.

    :param name:
    :return: name and required flag
    :rtype: tuple
    """
    required = False

    if name[-1] == '*':
        name = name[0:-1]
        required = True

    return name, required



"""
:mod:`zsl.task.job_context`
---------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from builtins import *

from typing import Callable
from abc import abstractmethod

from flask.wrappers import Response

from zsl.task.task_data import TaskData
from zsl.utils.reflection_helper import proxy_object_to_delegate


class Job(object):
    def __init__(self, data):
        # type: (dict) -> None

        self.data = data

    @property
    def path(self):
        """Job's path.

        :getter: Returns job's path
        :type: str
        """
        return self.data['path']

    @property
    def payload(self):
        """Data part of job.

        :getter: Returns job's payload
        :type: dict
        """
        return self.data['data']

    @property
    def is_valid(self):
        """Validity of job's data.

        :getter: Returns if job's data are valid
        :type: bool
        """
        return self.data and 'path' in self.data and 'data' in self.data


class JobContext(object):
    """Job Context"""

    _current_context = None

    def __init__(self, job, task, task_callable):
        # type: (Job, object, Callable) -> None
        """
        Constructor
        """
        self._job = job
        self._task = task
        self._task_callable = task_callable
        self._task_data = TaskData(self.job.payload)

        JobContext.set_current_context(None)

    @property
    def job(self):
        return self._job

    @property
    def task(self):
        return self._task

    @property
    def task_callable(self):
        return self._task_callable

    @property
    def task_data(self):
        return self._task_data

    @classmethod
    def get_current_context(cls):
        return cls._current_context

    @classmethod
    def set_current_context(cls, context):
        cls._current_context = context


class Responder(object):
    @abstractmethod
    def respond(self, r):
        pass


class StatusCodeResponder(Responder):
    def __init__(self, status_code):
        # type: (int)->None
        self._status_code = status_code

    def respond(self, r):
        # type: (Response)->None
        r.status_code = self._status_code


def add_responder(responder):
    # type:(Responder)->None
    jc = JobContext.get_current_context()
    if isinstance(jc, WebJobContext):
        jc.add_responder(responder)


def web_task_redirect(location):
    return {'headers': {'Location': location}, 'status_code': 301}


class WebJobContext(JobContext):
    def __init__(self, path, data, task, task_callable, request):
        """Constructor"""
        super(WebJobContext, self).__init__(Job({'data': data, 'path': path}), task, task_callable)
        self._request = request
        self._responders = []

    def get_web_request(self):
        return self._request

    def add_responder(self, r):
        self._responders.append(r)

    def notify_responders(self, response):
        try:
            for r in self._responders:
                r.respond(response)
        except:
            raise
        finally:
            self._responders = []


class DelegatingJobContext(JobContext):
    def __init__(self, job, task, task_callable, wrapped_job_context):
        super(DelegatingJobContext, self).__init__(job, task, task_callable)
        self._wrapped_job_context = wrapped_job_context
        proxy_object_to_delegate(self, wrapped_job_context)



"""
:mod:`zsl.task.task_data`
-------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from builtins import object

from zsl import inject, Injected, Zsl
from zsl.utils.warnings import deprecated


class TaskData(object):
    @inject(app=Zsl)
    def __init__(self, payload, app=Injected, payload_type=str):
        self._app = app
        self._payload = payload
        self._payload_type = payload_type

    @deprecated
    def get_data(self):
        return self._payload

    @property
    def payload(self):
        return self._payload

    def transform_payload(self, f):
        self._payload = f(self._payload) if self._payload is not None else {}



"""
:mod:`zsl.utils.nginx_push_helper`
----------------------------------

Helper for nginx push stream module
https://github.com/wandenberg/nginx-push-stream-module
"""
from __future__ import unicode_literals

from builtins import object
import requests
import json

from zsl.utils.url_helper import urlencode


class NginxPusher(object):
    def __init__(self, server_path, channel_prefix=None):
        self._server_path = server_path
        self._channel_prefix = (channel_prefix + '.') if channel_prefix is not None else ''

    def channel_path(self, channel_id):
        return '{0}?id={1}{2}'.format(self._server_path, self._channel_prefix, channel_id)

    def push_msg(self, channel_id, msg):
        """Push ``msg`` for given ``channel_id``. If ``msg`` is not string, it
         will be urlencoded
        """

        if type(msg) is not str:
            msg = urlencode(msg)

        return self.push(channel_id, msg)

    def push_object(self, channel_id, obj):
        """Push ``obj`` for ``channel_id``. ``obj`` will be encoded as JSON in
        the request.
        """

        return self.push(channel_id, json.dumps(obj).replace('"', '\\"'))

    def push(self, channel_id, data):
        """Push message with POST ``data`` for ``channel_id``
        """

        channel_path = self.channel_path(channel_id)
        response = requests.post(channel_path, data)

        return response.json()

    def delete_channel(self, channel_id):
        """Deletes channel
        """
        req = requests.delete(self.channel_path(channel_id))
        return req

        # TODO channel stats

        # TODO
        # def get_msg(self, channel_id):
        #    pass



"""
:mod:`zsl.tasks.asl.db_test_task`
---------------------------------

Created on 24.12.2012

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

from builtins import object
from injector import inject
import sqlalchemy.engine
from zsl import Zsl
from flask import Response


class DbTestTask(object):
    """
    Connects to a database and executes a simple query. The result of the query should be 6.

    <emph>Input</emph>
    No input.

    <emph>Output</emph>
    Returns just a number 6.

    @author: Martin Babka
    """

    @inject(db=sqlalchemy.engine.Engine, app=Zsl)
    def __init__(self, db, app):
        self._db = db
        self._app = app
        self._app.logger.debug("Call from DbTestTesk.__init__, db {0}".format(db))

    def perform(self, data):
        return Response("{0}".format(self._db.execute("select 1 * 2 * 3").scalar()), mimetype='text/plain')



"""
:mod:`zsl.application.initializers.service_initializer`
-------------------------------------------------------
"""
from __future__ import unicode_literals
from builtins import object

import importlib
import logging

from injector import singleton, Binder

from zsl import inject, Config, Injected
from zsl.utils.string_helper import camelcase_to_underscore


class ServiceInitializer(object):
    """Add outside services to application injector."""
    @staticmethod
    @inject(binder=Binder)
    def _bind_service(package_name, cls_name, binder=Injected):
        """Bind service to application injector.

        :param package_name: service package
        :type package_name: str
        :param cls_name: service class
        :type cls_name: str
        :param binder: current application binder, injected
        :type binder: Binder
        """
        module = importlib.import_module(package_name)
        cls = getattr(module, cls_name)

        binder.bind(
            cls,
            to=binder.injector.create_object(cls),
            scope=singleton
        )
        logging.debug("Created {0} binding.".format(cls))

    @staticmethod
    @inject(config=Config)
    def initialize(config):
        """Initialize method.

        :param config: current application config, injected
        :type config: Config
        """
        service_injection_config = config.get('SERVICE_INJECTION', ())

        if not isinstance(service_injection_config, (tuple, list)):
            service_injection_config = (service_injection_config,)

        for si_conf in service_injection_config:
            if isinstance(si_conf, str):
                package_name, cls_name = si_conf.rsplit('.', 1)
                ServiceInitializer._bind_service(package_name, cls_name)
            elif isinstance(si_conf, dict):
                services = si_conf['list']
                service_package = si_conf['package']

                for cls_name in services:
                    module_name = camelcase_to_underscore(cls_name)
                    package_name = "{0}.{1}".format(service_package, module_name)
                    ServiceInitializer._bind_service(package_name, cls_name)



"""
:mod:`zsl.application.initializers.library_initializer`
-------------------------------------------------------
"""
from __future__ import unicode_literals
from builtins import object

import logging
import sys
import os

from zsl import Config, inject


def append_paths(path, vendor_modules):
    new_path = []
    for v in vendor_modules:
        new_path.append(path + os.sep + v)
    sys.path = new_path + sys.path


class LibraryInitializer(object):
    """Add vendor modules to current path."""
    @staticmethod
    @inject(config=Config)
    def initialize(config):
        logging.debug("Initializing project external libraries.")

        external_libraries = config.get('EXTERNAL_LIBRARIES', None)
        if external_libraries is None:
            return

        vendor_path = external_libraries['vendor_path']
        append_paths(vendor_path, external_libraries['libs'])

        logging.info("Current PYTHON_PATH={0}.".format(sys.path))
        logging.debug("Project external libraries initialized.")



"""
Test resource policies.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

from unittest import TestCase
from typing import AnyStr
import sys

from zsl.resource.guard import ResourcePolicy, Access


def class_name(str_):
    # type: (str) -> AnyStr
    """Get name for `type` constructor with correct str type."""
    if sys.version_info.major == 2:
        return bytes(str_, 'utf-8')

    return str_


class TestResourcePolicy(TestCase):
    methods = ['create', 'read', 'update', 'delete']

    def testDefault(self):
        class TestPolicy(ResourcePolicy):
            pass

        policy = TestPolicy()

        self.assertEqual(Access.CONTINUE, policy.can_create__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.CONTINUE, policy.can_create__after(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.CONTINUE, policy.can_read__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.CONTINUE, policy.can_read__after(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.CONTINUE, policy.can_update__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.CONTINUE, policy.can_update__after(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.CONTINUE, policy.can_delete__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.CONTINUE, policy.can_delete__after(),
                         "should always return Access.CONTINUE")

    def testDefaultSetter(self):
        class TestPolicy(ResourcePolicy):
            default = Access.ALLOW

        policy = TestPolicy()

        self.assertEqual(Access.ALLOW, policy.can_create__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.ALLOW, policy.can_create__after(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.ALLOW, policy.can_read__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.ALLOW, policy.can_read__after(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.ALLOW, policy.can_update__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.ALLOW, policy.can_update__after(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.ALLOW, policy.can_delete__before(),
                         "should always return Access.CONTINUE")
        self.assertEqual(Access.ALLOW, policy.can_delete__after(),
                         "should always return Access.CONTINUE")

    def testCanProperty(self):
        for method in self.methods:
            # dynamically creates
            #
            # class TestPolicy(ResourcePolicy):
            #   can_'method' = Access.ALLOW
            #
            TestPolicy = type(class_name('TestPolicy'), (ResourcePolicy,),
                              {'can_' + method: Access.ALLOW})

            policy = TestPolicy()

            self.assertEqual(
                Access.ALLOW,
                getattr(policy, 'can_{}__before'.format(method))(),
                "should return Access.ALLOW for before {}".format(method)
            )
            self.assertEqual(
                Access.ALLOW,
                getattr(policy, 'can_{}__after'.format(method))(),
                "should return Access.ALLOW for after {}".format(method)
            )

            for m in filter(lambda x: x != method, self.methods):
                self.assertEqual(
                    Access.CONTINUE,
                    getattr(policy, 'can_{}__before'.format(m))(),
                    "should return Access.CONTINUE for before {}".format(method)
                )
                self.assertEqual(
                    Access.CONTINUE,
                    getattr(policy, 'can_{}__after'.format(m))(),
                    "should return Access.CONTINUE for after {}".format(method)
                )

    def testBeforeAndAfterMethods(self):
        for method in self.methods:
            for suffix in ['before', 'after']:
                # dynamically creates
                #
                # class TestPolicy(ResourcePolicy):
                #   def can_'method'__'suffix'(self, *args, **kwargs):
                #       return Access.ALLOW
                #
                TestPolicy = type(
                    class_name('TestPolicy'),
                    (ResourcePolicy,),
                    {
                        'can_{}__{}'.format(method, suffix):
                            lambda _: Access.ALLOW
                    }
                )

                policy = TestPolicy()

                self.assertEqual(
                    Access.ALLOW,
                    getattr(policy, 'can_{}__{}'.format(method, suffix))(),
                    "should return true for {} {}".format(method, suffix)
                )

                for m in self.methods:
                    for s in ['before', 'after']:
                        if m == method and s == suffix:
                            continue

                        self.assertEqual(
                            Access.CONTINUE,
                            getattr(policy, 'can_{}__{}'.format(m, s))(),
                            "should return Access.CONTINUE for {} and {}"
                            .format(m, s)
                        )



"""
:mod:`zsl.utils.security_helper`
-------------------------------------

Helper module with function dealing with security.
"""

from __future__ import unicode_literals
from builtins import str

from flask.config import Config

from zsl import inject
import hashlib
from zsl.utils.string_helper import generate_random_string

TOKEN_RANDOM = 'random_token'
TOKEN_HASHED = 'hashed_token'

TOKEN_SERVICE_SECURITY_CONFIG = 'SERVICE_SECURITY_TOKEN'


def generate_security_data():
    """Generate security token - a random token and its hashed version salted
    with a secret.

    :return: random and hashed token
    :rtype: dict(str, str)
    """
    random_token = generate_random_string()
    return {TOKEN_RANDOM: random_token, TOKEN_HASHED: compute_token(random_token)}


def wrap_plain_data_as_secure(data):
    """Wrap task data with security token.

    :param data: data to be wrapped
    :return: wrapped data with security token
    :rtype: dict
    """
    return {'security': generate_security_data(), 'data': data}


@inject(config=Config)
def compute_token(random_token, config):
    """Compute a hash of the given token with a preconfigured secret.

    :param random_token: random token
    :type random_token: str
    :return: hashed token
    :rtype: str
    """
    secure_token = config[TOKEN_SERVICE_SECURITY_CONFIG]
    sha1hash = hashlib.sha1()
    sha1hash.update(random_token + secure_token)
    return sha1hash.hexdigest().upper()


def verify_security_data(security):
    """Verify an untrusted security token.

    :param security: security token
    :type security: dict
    :return: True if valid
    :rtype: bool
    """
    random_token = security[TOKEN_RANDOM]
    hashed_token = security[TOKEN_HASHED]
    return str(hashed_token) == str(compute_token(random_token))



"""
:mod:`zsl.utils.string_helper`
------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from builtins import str
from builtins import range
import re
import random
import string

_html_tag_re = re.compile(r'''</?\w+((\s+\w+(\s*=\s*(?:".*?"|'.*?'|[^'">\s]+))?)+\s*|\s*)/?>''', flags=re.IGNORECASE)
_html_tag_re_un = re.compile(
    r'''</?\w+((\s+\w+(\s*=\s*(?:".*?"|'.*?'|[^'">\s]+))?)+\s*|\s*)/?>''', flags=re.IGNORECASE | re.UNICODE)


def underscore_to_camelcase(value, first_upper=True):
    """Transform string from underscore_string to camelCase.

    :param value: string with underscores
    :param first_upper: the result will have its first character in upper case
    :type value: str
    :return: string in CamelCase or camelCase according to the first_upper
    :rtype: str

    :Example:
        >>> underscore_to_camelcase('camel_case')
        'CamelCase'
        >>> underscore_to_camelcase('camel_case', False)
        'camelCase'
    """
    value = str(value)
    camelized = "".join(x.title() if x else '_' for x in value.split("_"))
    if not first_upper:
        camelized = camelized[0].lower() + camelized[1:]
    return camelized


def camelcase_to_underscore(name):
    """Transform string from camelCase to underscore_string.

    :param name: string in camelcase
    :type name: str
    :return: string with underscores
    :rtype: str

    :Example:
        >>> camelcase_to_underscore('camelCase')
        'camel_case'
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def et_node_to_string(et_node, default=''):
    """Simple method to get stripped text from node or ``default`` string if None is given.

    :param et_node: Element or None
    :param default: string returned if None is given, default ``''``
    :type et_node: xml.etree.ElementTree.Element, None
    :type default: str
    :return: text from node or default
    :rtype: str
    """

    return str(et_node.text).strip() if et_node is not None and et_node.text else default


def generate_random_string(size=6, chars=string.ascii_uppercase + string.digits):
    """Generate random string.

    :param size: Length of the returned string. Default is 6.
    :param chars: List of the usable characters. Default is string.ascii_uppercase + string.digits.
    :type size: int
    :type chars: str
    :return: The random string.
    :rtype: str
    """
    return ''.join(random.choice(chars) for _ in range(size))


def addslashes(s, l=None):
    """Add slashes for given characters. Default is for ``\`` and ``'``.

    :param s: string
    :param l: list of characters to prefix with a slash ``\``
    :return: string with slashed characters
    :rtype: str

    :Example:
        >>> addslashes("'")
        "\\'"
    """
    if l is None:
        l = ["\\", "'", ]

    # l = ["\\", '"', "'", "\0", ]
    for i in l:
        if i in s:
            s = s.replace(i, '\\' + i)
    return s


def xstr(s):
    """If ``s`` is None return empty string.

    :param s: string
    :return: s or an empty string if s is None
    :rtype: str

    :Example:
        >>> xstr(None)
        ''
    """
    return '' if s is None else str(s)


def xunicode(s):
    """If ``s`` is None return empty string

    .. deprecated::
       Use :func:`xstr` instead.

    :param s: string
    :return: s or an empty string if s in None
    :rtype: str
    """
    return '' if s is None else str(s)


def strip_html_tags(s):
    """Remove all html tags from string.

    :param s: string
    :type s: str
    :return: string with stripped html tags
    :rtype: str

    :Example:
        >>> strip_html_tags('<h1>hello <i>world</i></h1>')
        'hello world'
    """
    return _html_tag_re.sub('', s)


def strip_html_tags_unicode(s):
    """Remove all html tags from string.

    .. deprecated::
       Use :func:`strip_html_tags` instead.

    :param s: string
    :type s: str
    :return: string with stripped html tags
    :rtype: str
    """
    return _html_tag_re_un.sub('', s)



"""
:mod:`zsl.utils.reflection_helper`
----------------------------------

Helper module for OOP operations.

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals


def extend(instance, new_class):
    """Adds new_class to the ancestors of instance.

    :param instance: Instance that will have a new ancestor.
    :param new_class: Ancestor.
    """
    instance.__class__ = type(
        '%s_extended_with_%s' % (instance.__class__.__name__, new_class.__name__),
        (new_class, instance.__class__,),
        {}
    )


def add_mixin(base_class, mixin_class):
    base_class.__bases__ = (mixin_class,) + base_class.__bases__
    return base_class


def is_scalar(v):
    return isinstance(v, (type(None), str, int, float, bool))


def proxy_object_to_delegate(proxy_object, delegate_object):
    proxy_class_name = 'Proxy{0}To{1}'.format(
        proxy_object.__class__.__name__,
        delegate_object.__class__.__name__)
    proxy_object.__class__ = type(proxy_class_name,
                                  proxy_object.__class__.__bases__,
                                  dict(proxy_object.__class__.__dict__))
    proxy_object.__class__.__bases__ = (delegate_object.__class__,)
    proxy_object._delegate_object = delegate_object

    def __getattr__(self, item):
        return getattr(self._delegate_object, item)

    proxy_object.__class__.__getattr__ = __getattr__



"""
:mod:`zsl.utils.cache_helper`
-----------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from builtins import object
from zsl import inject
from zsl.cache.id_helper import IdHelper, model_key_generator, create_key_object_prefix,\
    app_model_decoder_fn, app_model_encoder_fn
from zsl.db.model.app_model_json_encoder import AppModelJSONEncoder
import json
import abc
import logging
from zsl.task.job_context import JobContext


class CacheDecorator(object):

    @inject(id_helper=IdHelper)
    def __init__(self, id_helper):
        self._id_helper = id_helper
        logging.debug("Initializing CacheDecorator.")

    def decorate(self, key_params, timeout, fn):
        self._key_params = key_params
        self._fn = fn
        self._timeout = timeout

        return self.get_wrapped_fn()

    @staticmethod
    def is_caching(*args):
        jc = JobContext.get_current_context()
        if 'cache' not in jc.job.data:
            cs = True  # TODO: Default
        else:
            cs = bool(jc.job.data['cache'])

        logging.debug("Cache status: '%s'.", cs)
        return cs

    @abc.abstractmethod
    def get_wrapped_fn(self):
        pass

    def get_data_key(self, *args):
        prefix = create_key_object_prefix(args[0])
        key = create_key_for_data(prefix, args[1], self._key_params)
        return key

    @staticmethod
    def get_decoder():
        return app_model_decoder_fn

    @staticmethod
    def get_encoder():
        return app_model_encoder_fn


class CacheOutputDecorator(CacheDecorator):

    def get_wrapped_fn(self):
        def wrapped_fn(*args):
            if not self.is_caching(*args):
                return self._fn(*args)

            key = self.get_data_key(*args)
            logging.debug("Initializing CacheOutputDecorator - key %s.", key)

            if self._id_helper.check_key(key):
                logging.debug("Retrieved from cache.")
                return self._id_helper.get_key(key)
            else:
                ret_val = self._fn(*args)
                if not isinstance(ret_val, (str, bytes)):
                    raise Exception("Can not cache non-string value. Is the serialization, json_output, already done?")

                self._id_helper.set_key(key, ret_val, self._timeout)
                logging.debug("Newly fetched into the cache.")
                return ret_val

        return wrapped_fn


def cache_output(key_params, timeout='default'):
    def decorator_fn(fn):
        return CacheOutputDecorator().decorate(key_params, timeout, fn)

    return decorator_fn


class CacheModelDecorator(CacheDecorator):

    def get_wrapped_fn(self):
        def wrapped_fn(*args):
            if not self.is_caching(*args):
                return self._fn(*args)

            key = self.get_data_key(*args)
            logging.debug("Initializing CacheModelDecorator - key %s.", key)

            if self._id_helper.check_key(key):
                model_key = self._id_helper.get_key(key)
                logging.debug("Retrieved from cache %s.", model_key)

                if self._id_helper.check_key(model_key):
                    return self.get_decoder()(model_key, self._id_helper.get_key(model_key))

            model = self._fn(*args)
            if model is None:
                return model
            encoded_model = json.dumps(model, cls=AppModelJSONEncoder)
            model_key = model_key_generator(model)
            self._id_helper.set_key(key, model_key, self._timeout)
            self._id_helper.set_key(model_key, encoded_model, self._timeout)
            logging.debug("Newly fetched into the cache.")
            return model

        return wrapped_fn


def cache_model(key_params, timeout='default'):
    """
    Caching decorator for app models in task.perform
    """
    def decorator_fn(fn):
        return CacheModelDecorator().decorate(key_params, timeout, fn)

    return decorator_fn


class CachePageDecorator(CacheDecorator):

    def get_wrapped_fn(self):
        def wrapped_fn(*args):
            if not self.is_caching(*args):
                return self._fn(*args)

            page_key = self.get_data_key(*args)
            logging.debug("Initializing CachePageDecorator - key %s.", page_key)

            if self._id_helper.check_page(page_key):
                logging.debug("Retrieved from cache %s.", page_key)
                return self._id_helper.gather_page(page_key, self.get_decoder())

            page = self._fn(*args)
            self._id_helper.fill_page(page_key, page, self._timeout, self.get_encoder())
            logging.debug("Newly fetched into the cache.")
            return page

        return wrapped_fn


def cache_page(key_params, timeout='default'):
    """
    Cache a page (slice) of a list of AppModels
    """
    def decorator_fn(fn):
        d = CachePageDecorator()
        return d.decorate(key_params, timeout, fn)

    return decorator_fn

# alias
cache_list = cache_page


def cache_filtered_page(filter_param='filter', pagination_param='pagination', sorter_param='sorter', timeout='default'):
    def decorator_fn(fn):
        d = CachePageDecorator()
        return d.decorate([filter_param, sorter_param, pagination_param], timeout, fn)

    return decorator_fn


def create_key_for_data(prefix, data, key_params):
    """
    From ``data`` params in task create corresponding key with help of ``key_params`` (defined in decorator)
    """
    d = data.get_data()
    values = []
    for k in key_params:
        if k in d and type(d[k]) is list:
            values.append("{0}:{1}".format(k, " -".join(d[k])))
        else:
            value = d[k] if k in d else ''
            values.append("{0}:{1}".format(k, value))

    return "{0}-{1}".format(prefix, "-".join(values))



"""
:mod:`zsl.db.model.app_model_json_decoder`
------------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from json.decoder import JSONDecoder, WHITESPACE
from zsl.utils.import_helper import fetch_class


def get_json_decoder(full_class_name, hints=None):
    class AppModelJSONDecoder(JSONDecoder):
        def decode(self, s, _w=WHITESPACE.match):
            values = JSONDecoder.decode(self, s, _w=_w)
            model = fetch_class(full_class_name)(values, 'id', hints)
            return model

    return AppModelJSONDecoder



"""
Test http responses of resource guard. It should return a model if policy is
met and an proper http code when policy is broken.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *
from unittest.case import TestCase

from mocks import mock

import http.client

from zsl import Zsl
from zsl.application.containers.web_container import WebContainer
from zsl.interface.resource import ResourceResult
from zsl.resource.guard import (guard, GuardedMixin, Access,
                                PolicyViolationError, ResourcePolicy)

from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.http import json_loads, HTTPTestCase
from zsl.testing.test_utils import parent_module

TEST_VALUE_CREATED = 'created'
TEST_VALUE_READ = 'read'
TEST_VALUE_UPDATED = 'updated'
TEST_VALUE_DELETED = 'deleted'

SAMPLE_MODEL = {
    'test': 'value',
    'one': 1
}


class GuardedResourceTestResource(object):
    """This will be patched."""
    pass


class GuardedResourceTest(TestCase, HTTPTestCase):
    PATH = '/resource/guarded_resource_test'
    RESOURCE_CLASS = 'resource.guarded_resource_test.GuardedResourceTestResource'

    def setUp(self):
        config_object = IN_MEMORY_DB_SETTINGS.copy()
        # add this package as resource package for zsl to find the
        # `JsonServerModelResourceResource`
        config_object['RESOURCE_PACKAGES'] = ('resource',)
        zsl = Zsl(__name__, config_object=config_object,
                  modules=WebContainer.modules())
        zsl.testing = True

        # mock http requests
        self.app = zsl.test_client()

    def testAllowPolicy(self):
        class AllowPolicy(ResourcePolicy):
            default = Access.ALLOW

        @guard([AllowPolicy()])
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')

            self.assertHTTPStatus(
                http.client.OK,
                rv.status_code,
                "should return 200 status, returned data {0}".format(rv.data)
            )
            data = json_loads(rv.data)
            self.assertDictEqual(SAMPLE_MODEL, data,
                                 "should return sample model")

    def testDefaultDenyPolicy(self):
        class DenyPolicy(ResourcePolicy):
            default = Access.DENY

        @guard([DenyPolicy()])
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')
            data = json_loads(rv.data)

            self.assertDictEqual({}, data, "should return an empty model")
            self.assertHTTPStatus(http.client.FORBIDDEN, rv.status_code,
                                  "should return default 403 status")

    def testCustomDenyException(self):
        class PaymentError(PolicyViolationError):
            def __init__(self):
                super(PaymentError, self).__init__('Payment required', code=402)

        class PaymentPolicy(ResourcePolicy):
            @property
            def default(self):
                raise PaymentError()

        @guard([PaymentPolicy()])
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')
            data = json_loads(rv.data)

            self.assertDictEqual({}, data, "should return an empty model")
            self.assertHTTPStatus(http.client.PAYMENT_REQUIRED, rv.status_code,
                                  "should return custom status 402")

    def testCustomErrorHandling(self):
        def error_handler(*_):
            return ResourceResult(body={}, status=301)

        class DenyPolicy(ResourcePolicy):
            default = Access.DENY

        @guard(
            policies=[DenyPolicy()],
            exception_handlers=[error_handler]
        )
        class Resource(GuardedMixin):
            def read(self, *args, **kwargs):
                return SAMPLE_MODEL

        with mock.patch(self.RESOURCE_CLASS, Resource):
            rv = self.app.get(self.PATH + '/1')
            data = json_loads(rv.data)

            self.assertDictEqual({}, data, "should return an empty model")
            self.assertHTTPStatus(http.client.MOVED_PERMANENTLY, rv.status_code,
                                  "should return custom status 301")




"""
:mod:`zsl.application.containers.gearman_container`
---------------------------------------------------
"""
from __future__ import unicode_literals

from zsl.application.containers.core_container import CoreContainer
from zsl.application.modules.gearman_module import GearmanModule


class GearmanContainer(CoreContainer):
    """Configuration for Gearman application."""
    worker = GearmanModule



from unittest.case import TestCase

from zsl.utils.string_helper import camelcase_to_underscore, underscore_to_camelcase


class InflectionTestCase(TestCase):
    def testCamelCaseToUnderscore(self):
        self.assertEquals(
            "camel_case_to_underscore",
            camelcase_to_underscore("camelCaseToUnderscore"),
            "CC to usc conversion"
        )
        self.assertEquals(
            "camel_case_to_underscore",
            camelcase_to_underscore("CamelCaseToUnderscore"),
            "CC to usc conversion"
        )

    def testUnderscoreToCamelCase(self):
        self.assertEquals(
            "camelCaseToUnderscore",
            underscore_to_camelcase("camel_case_to_underscore", False),
            "CC to usc conversion"
        )
        self.assertEquals(
            "CamelCaseToUnderscore",
            underscore_to_camelcase("camel_case_to_underscore"),
            "CC to usc conversion"
        )
        self.assertEquals(
            "CamelCaseToUnderscore",
            underscore_to_camelcase("camel_case_to_underscore", True),
            "CC to usc conversion"
        )


"""
:mod:`zsl.db.helpers.sorter`
----------------------------
"""
from __future__ import unicode_literals
from builtins import zip
from builtins import object
from sqlalchemy import desc, asc

DEFAULT_SORT_ORDER = 'ASC'  # If changed, look at the condition in apply_sorter if self.get_order() == "DESC":.


class Sorter(object):
    """
    Helper class for applying ordering criteria to query.
    """

    def __init__(self, sorter, mappings=None):
        """
        sorter = {'sortby': string, 'sort': string}
            sortby - string of comma-separated column names by which you want to order
            sort - string of comma-separated values 'ASC'/'DESC' (order direction) which
                    set order direction to corresponding columns from sorter['sortby'] string
            notes: - if 'sortby' key is not in sorter, no sorting will be applied to query
                   - if 'sort' key is not in sorter, DEFAULT_SORT_ORDER will be applied to
                     all columns from sorter['sortby']
                   - if sorter['sort'] == 'ASC' / 'DESC' (contains only one order direction),
                     this direction will be applied to all columns from sorter['sortby']
                   - if you want to order only by one column, simply put
                     sorter['sortby'] = '<column_name>' - without comma at the end of
                     the string
        mappings dict - maps column names from sorter['sortby'] to column attributes names of
                        objects (see example)
                      - if the column names from sorter['sortby'] is equal to the name
                        of column attribute, it doesn`t have to be mentioned in mappings

        Example:
            sorter = {'sortby': 'firstname,state,sport', 'sort': 'ASC'}

            mappings = {
                'state': (State, 'name_sk'),
                'sport': (Sport, 'name'),
            }

        """
        if mappings is None:
            mappings = []

        if 'sortby' in sorter:
            self._fields = sorter['sortby'].split(',')

            if 'sort' in sorter:
                self._orders = sorter['sort'].split(',')
                if len(self._orders) == 1:
                    self._orders *= len(self._fields)
                elif len(self._orders) != len(self._fields):
                    raise Exception(
                        'zsl.db.helpers.Sorter: Number of order settings is nor zero nor one nor equal to number of'
                        'sortby columns.')

            else:
                self._orders = [DEFAULT_SORT_ORDER] * len(self._fields)

            self._enabled = True
        else:
            self._enabled = False

        self._mappings = mappings

    def is_enabled(self):
        return self._enabled

    def get_fields(self):
        return self._fields

    def get_orders(self):
        return self._orders

    def apply_sorter(self, q, cls):
        if self.is_enabled():
            sorter_settings = []
            for field, order in zip(self.get_fields(), self.get_orders()):
                if field in self._mappings:
                    (cls, mapped_field) = self._mappings[field]
                    attr = getattr(cls, mapped_field)
                else:
                    attr = getattr(cls, field)
                if order == "DESC":  # If changed, look at the DEFAULT_SORT_ORDER definition.
                    sorter_settings.append(desc(attr))
                else:
                    sorter_settings.append(asc(attr))

            return q.order_by(*sorter_settings)
        else:
            return q



# coding: utf-8
"""
:mod:`zsl.utils.sk_validators_helper`
-------------------------------------

Helper module for user form validation, in Slovak.
"""
# TODO make more general.

from __future__ import unicode_literals
from builtins import object
import wtforms.validators
from time import strptime


# from symbol import except_clause

# TODO document this, ask Martin or Jan Janco
class Validators(object):
    class Optional(wtforms.validators.Optional):
        def __init__(self, strip_whitespace=True):
            wtforms.validators.Optional.__init__(self, strip_whitespace)

    class Required(wtforms.validators.Required):
        def __init__(self, message='Toto pole je povinné.'):
            wtforms.validators.Required.__init__(self, message)

    class URL(wtforms.validators.URL):
        def __init__(self, require_tld=True,
                     message='Neplatná URL adresa. URL adresa musí začínať s "http://". Napríklad'
                             '"http://www.zoznam.sk"'):
            wtforms.validators.URL.__init__(self, require_tld, message)

    class Date(object):
        """
        validator - skontroluje, ci dany string je datum v zadanom formate a ci je dany datum platny
        """

        def __init__(self, format, message='Nesprávny formát dátumu alebo neplatný dátum.', stop_after_fail=False):
            self.format = format
            self.message = message
            self.stop_after_fail = stop_after_fail

        def __call__(self, form, field):
            try:
                strptime(field.data, self.format)
            except ValueError:
                if self.stop_after_fail:
                    raise wtforms.validators.StopValidation(self.message)
                else:
                    raise wtforms.validators.ValidationError(self.message)

    class Length(wtforms.validators.Length):
        def __init__(self, min=-1, max=-1, message=None):
            if message:
                msg = message
            else:
                msg = ''
                if min > -1:
                    msg += 'Minimálny počet znakov textu: %(min)d. '
                if max > -1:
                    msg += 'Maximálny počet znakov textu: %(max)d.'

            wtforms.validators.Length.__init__(self, min, max, msg)

    class Regexp(wtforms.validators.Regexp):
        def __init__(self, regex, flags=0, message='Nesprávny formát dát.'):
            wtforms.validators.Regexp.__init__(self, regex, flags, message)



"""
:mod:`zsl.interface.webservice.importer`
----------------------------------------
"""
from __future__ import unicode_literals
import sys
import os

# TODO: Consider removing automatic path initialization!
sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..', '..'))



"""
:mod:`zsl.testing.zsl`
----------------------
This module allows for unit testing with a Zsl instance. Use
:class:`.ZslTestCase` as a mixin to create a predefined Zsl instance to be used
while testing. Refer to unit testing section :ref:`unit-testing-zsl-instance`
for an example.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import *

from collections import namedtuple

import logging
from functools import partial

from zsl import Zsl
from zsl._state import set_current_app
from zsl.application.service_application import set_profile
from zsl.configuration import InvalidConfigurationException

ZslTestConfiguration = namedtuple('ZslTestConfiguration',
                                  ['app_name', 'version', 'container',
                                   'profile', 'config_object'])
ZslTestConfiguration.__new__ = partial(ZslTestConfiguration.__new__,
                                       config_object=None,
                                       version=None,
                                       profile=None)


class ZslTestCase(object):
    ZSL_TEST_CONFIGURATION = None  # type: ZslTestConfiguration

    @classmethod
    def setUpClass(cls):
        if cls.ZSL_TEST_CONFIGURATION is None:
            raise InvalidConfigurationException("Please give a test container "
                                                "specification via 'container' "
                                                "class variable.")

        config = cls.ZSL_TEST_CONFIGURATION  # type: ZslTestConfiguration
        if config.profile:
            set_profile(config.profile)
        app = Zsl(config.app_name + "-test",
                  version=config.version,
                  modules=config.container.modules(),
                  config_object=config.config_object)
        logging.getLogger(config.app_name).debug(
            "ZSL test app created {0}.".format(app))

        super(ZslTestCase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        config = cls.ZSL_TEST_CONFIGURATION  # type: ZslTestConfiguration
        logging.getLogger(config.app_name).debug(
            "ZSL test app tear down {0}.".format(config.app_name))
        set_current_app(None)



"""
:mod:`zsl.application.modules.context_module`
---------------------------------------------

Application context describes the environment in which the application is run\
and is responsible for specific initializations on it.
"""
from __future__ import unicode_literals

import logging

from zsl.application.initialization_context import InitializationContext
from injector import singleton, Module

from zsl.application.initializers.library_initializer import LibraryInitializer
from zsl.application.initializers.service_initializer import ServiceInitializer
from zsl.application.initializers.unittest_initializer import UnitTestInitializer

#: Initializers used in all applications
default_initializers = (LibraryInitializer, ServiceInitializer)

#: Initializers used in unit test applications
unittest_initializers = default_initializers + (UnitTestInitializer,)


class DefaultContextModule(Module):
    """Adds default application context to current configuration."""

    def _create_context(self):
        logging.getLogger(__name__).debug("Creating default context.")
        return InitializationContext(initializers=default_initializers)

    def configure(self, binder):
        context = self._create_context()
        binder.bind(InitializationContext, to=context, scope=singleton)


class WorkerContextModule(DefaultContextModule):
    """Worker application context to current configuration."""
    pass


class TestContextModule(DefaultContextModule):
    """Test application context to current configuration"""

    def _create_context(self):
        return InitializationContext(unit_test=True, initializers=unittest_initializers)



"""
:mod:`zsl.utils.deploy.js_model_generator`
------------------------------------------

.. moduleauthor:: Peter Morihladko
"""
from __future__ import unicode_literals

from builtins import range
from builtins import object

import sys
import json
import importlib
import hashlib
from typing import Union

from sqlalchemy.orm import class_mapper
import sqlalchemy.exc

from zsl.utils.string_helper import camelcase_to_underscore, underscore_to_camelcase
from zsl.utils.deploy.integrator import integrate_to_file

model_tpl = """    {model_prefix}{model_name} = {model_fn}.extend({{
        urlRoot: App.service_url + 'resource/{resource_name}',
        schema: {schema}
    }});

    {collection_prefix}{model_name} = {collection_fn}.extend({{
        model: {model_prefix}{model_name},
        url: {model_prefix}{model_name}.prototype.urlRoot
    }});
"""

list_opts_tpl = """function(callback, field) {{
                    field.setOptions(new {collection_prefix}{model_name}([],{{limit: 'unlimited'}}));
                }}"""


class ModelGenerator(object):
    def __init__(self, module, model_prefix="", collection_prefix="", model_fn="Atteq.bb.Model",
                 collection_fn="Atteq.bb.Collection"):
        self.model_prefix = model_prefix
        self.collection_prefix = collection_prefix
        self.model_fn = model_fn
        self.collection_fn = collection_fn
        self.table_to_class = {}

        self.models = importlib.import_module(module)

    def _get_list_options(self, column):
        fk = list(column.foreign_keys)[0]

        table_name = underscore_to_camelcase(fk.column.table.name)

        return list_opts_tpl.format(collection_prefix=self.collection_prefix, model_name=table_name)

    def _map_table_name(self, model_names):
        """
        Pre foregin_keys potrbejeme pre z nazvu tabulky zistit class,
        tak si to namapujme
        """

        for model in model_names:
            if isinstance(model, tuple):
                model = model[0]

            try:
                model_cls = getattr(self.models, model)
                self.table_to_class[class_mapper(model_cls).tables[0].name] = model
            except AttributeError:
                pass

    def generate_model(self, model_name, model_plural=None):
        if model_name not in dir(self.models):
            raise ImportError(
                "Model [{name}] couldn't be found in {module}\n".format(name=model_name, module=self.models.__name__))

        if model_plural is None:
            model_plural = model_name + 's'

        model = getattr(self.models, model_name)

        schema = {}
        mapper = class_mapper(model)
        callbacks = []

        for column in mapper.columns:
            col_type = column.type.__class__.__name__
            attrs = {}

            if column.primary_key:
                continue

            if column.foreign_keys:
                try:
                    attrs['type'] = 'AtteqSelect'
                    attrs['options'] = '__CALLBACK__%d' % len(callbacks)
                    callbacks.append(self._get_list_options(column))

                    # TODO uf uf uuuuf
                    fk_table = list(column.foreign_keys)[0].target_fullname.split('.')[0]

                    if fk_table in self.table_to_class:
                        attrs['foreign_model'] = '%s%s' % (self.model_prefix, self.table_to_class[fk_table])

                except sqlalchemy.exc.NoReferencedTableError:
                    attrs['type'] = 'Text'

            elif col_type == 'TEXT':
                attrs['type'] = "TextArea"

            elif col_type == 'Enum':
                attrs['type'] = 'AtteqSelect' if column.nullable else 'Select'
                attrs['options'] = column.type.enums

            elif col_type == 'INTEGER':
                attrs['type'] = 'Number'

            else:
                attrs['type'] = "Text"

            if column.nullable:
                attrs['nullable'] = True

            schema[column.name] = attrs

        schema = "\n        ".join(json.dumps(schema, indent=4).split("\n"))

        for i in range(len(callbacks)):
            schema = schema.replace('"__CALLBACK__%d"' % i, callbacks[i])

        return model_tpl.format(
            model_name=model_name,
            model_prefix=self.model_prefix,
            collection_prefix=self.collection_prefix,
            resource_name=camelcase_to_underscore(model_plural),
            model_fn=self.model_fn,
            collection_fn=self.collection_fn,
            schema=schema
        )

    def generate_models(self, models):
        js_models = []

        self._map_table_name(models)

        for model in models:
            if isinstance(model, tuple):
                model_name = model[0]
                model_plural = model[1]
            else:
                model_name = model
                model_plural = None

            js_model = self.generate_model(model_name, model_plural)
            js_models.append(js_model)

        return js_models


def parse_model_arg(models):
    # type: (list[str]) -> list[str]
    """Parse the model argument definition."""
    return [tuple(m.split('/')) if '/' in m else m for m in models]


def generate_js_models(module, models, collection_prefix, model_prefix,
                       model_fn, collection_fn, marker, integrate, js_file):
    # type: (str, str, str, str, str, str, str, bool, str) -> Union[str, None]
    """Generate models for Backbone Javascript applications.

    :param module: module from which models are imported
    :param models: model name, can be a tuple WineCountry/WineCountries as singular/plural
    :param model_prefix: namespace prefix for models (app.models.)
    :param collection_prefix: namespace prefix for collection (App.collections.)
    :param model_fn: name of model constructor (MyApp.bb.Model)
    :param collection_fn: name of collection constructor (MyApp.bb.Collection)
    :param marker: marker to indicate the auto generated code
    :param integrate: integrate to file
    :param js_file: file to integrate
    :return: generated models or nothing if writing into a file
    """

    options = {
        'model_prefix': model_prefix,
        'collection_prefix': collection_prefix,
        'model_fn': model_fn,
        'collection_fn': collection_fn
    }

    generator = ModelGenerator(module,
                               **{o: options[o] for o in options if options[o] is not None})

    models = generator.generate_models(parse_model_arg(models))

    if integrate:
        sys.stderr.write("Integrate is really experimental")

        if not marker:
            marker = hashlib.md5("{0}{1}".format(module, models)).hexdigest()

        start = "// * -- START AUTOGENERATED %s -- * //\n" % marker
        end = "// * -- END AUTOGENERATED %s -- * //\n" % marker

        return integrate_to_file("\n".join(models), js_file, start, end)

    else:
        return "\n".join(models)



"""
:mod:`zsl.interface.gearman.worker`
-----------------------------------

.. moduleauthor:: Martin
"""
from __future__ import unicode_literals
import gearman
from gearman.job import GearmanJob

from zsl.interface.task_queue import TaskQueueWorker
from zsl.task.job_context import Job
from zsl.interface.gearman.json_data_encoder import JSONDataEncoder


class ReloadingWorker(gearman.GearmanWorker):
    def __init__(self, host_list=None):
        super(ReloadingWorker, self).__init__(host_list)
        self._should_stop = False

    def on_job_complete(self, current_job, job_result):
        super(ReloadingWorker, self).on_job_complete(current_job, job_result)
        if self._should_stop:
            quit()
        return True


def job_from_gearman_job(gearman_job):
    # type: (GearmanJob) -> Job
    """Creates zsl job from gearman job.

    :param gearman_job: gearman job
    :type gearman_job: GearmanJob
    :return: zsl job
    :rtype: Job
    """

    return Job(gearman_job.data)


class GearmanTaskQueueWorker(TaskQueueWorker):
    def __init__(self):
        super(GearmanTaskQueueWorker, self).__init__()

        self.gearman_worker = ReloadingWorker(
            ["{0}:{1}".format(self._config['GEARMAN']['host'],
                              self._config['GEARMAN']['port'])])
        self.gearman_worker.set_client_id(self._get_client_id())
        self.gearman_worker.data_encoder = JSONDataEncoder
        self.gearman_worker.register_task(self._config['GEARMAN_TASK_NAME'], self.execute_gearman_job)
        self.gearman_worker.logical_worker = self

        self._current_worker = None

    def stop_worker(self):
        self._app.logger.info("Stopping Gearman worker on demand - quitting.")
        self._current_worker._should_stop = True

    def execute_gearman_job(self, worker, job):
        # type: (ReloadingWorker, GearmanJob) -> dict
        job = job_from_gearman_job(job)
        self._current_worker = worker

        return self.execute_job(job)

    def run(self):
        self._app.logger.info("Running the worker.")
        self.gearman_worker.work()



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *

from unittest.case import TestCase

from zsl.interface.task import create_simple_model


class ParentClass(object):
    pass


class CreateSimpleModelTestCase(TestCase):
    def testCreateSimpleModelWithDefaults(self):
        point = create_simple_model('Point', ['x', 'y'], {'x': 0.0, 'y': 1.0})
        p = point()
        self.assertIsInstance(p, object, "Point must be instance of object.")
        self.assertEqual(p.x, 0.0, "Zero is default for x")
        self.assertEqual(p.y, 1.0, "One is default for y")

    def testCreateSimpleModelWithParent(self):
        point = create_simple_model('Point', ['x', 'y'], parent=ParentClass)
        p = point()
        self.assertIsInstance(p, ParentClass,
                              "Point must be instance of ParentClass.")

    def testCreateSimpleModelWithModule(self):
        module_mock = 0
        point = create_simple_model('Point', ['x', 'y'], parent=ParentClass,
                                    model_module=module_mock)
        p = point()
        self.assertEqual(module_mock, p.__module__)
        self.assertIsInstance(p, ParentClass,
                              "Point must be instance of ParentClass.")

    def testCreateSimpleModelWithoutDefaults(self):
        point = create_simple_model('Point', ['x', 'y'])
        p = point()
        self.assertIsNone(p.x, "None is default")
        self.assertIsNone(p.y, "None is default")



"""
:mod:`zsl.application.modules.gearman_module`
---------------------------------------------
"""
from __future__ import unicode_literals

import click
from injector import Module, ClassProvider, singleton

from zsl import Zsl, inject
from zsl.application.modules.cli_module import ZslCli
from zsl.interface.gearman.task_filler import exec_task_filler
from zsl.interface.task_queue import TaskQueueWorker, run_worker
from zsl.interface.gearman.worker import GearmanTaskQueueWorker
from zsl.utils.injection_helper import simple_bind


class GearmanCli(object):
    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli)->GearmanCli

        @zsl_cli.cli.group("Gearman related tasks.")
        def gearman():
            pass

        @gearman.command(help="run worker")
        def worker():
            # type: () -> None
            run_worker()

        @gearman.command()
        @click.argument('task_path', metavar='task')
        @click.argument('data', default=None, required=False)
        def task_filler(task_path, data):
            exec_task_filler(task_path, data)

        self._gearman = gearman

    @property
    def gearman(self):
        return self._gearman


class GearmanModule(Module):
    """Adds gearman to current configuration."""

    def configure(self, binder):
        binder.bind(TaskQueueWorker, to=ClassProvider(GearmanTaskQueueWorker), scope=singleton)
        simple_bind(binder, GearmanCli, singleton)



# coding: utf-8
"""
:mod:`zsl.utils.url_helper`
---------------------------

Helper module for URL handling.
"""
from __future__ import unicode_literals
from builtins import str
import unicodedata
import re
import urllib


def slugify(value, allow_unicode=False):
    """Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    :param value: string
    :param allow_unicode: allow utf8 characters
    :type allow_unicode: bool
    :return: slugified string
    :rtype: str

    :Example:
        >>> slugify('pekná líščička')
        'pekna-liscicka'
    """
    value = str(value)

    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
        value = re.sub(r'[^\w\s-]', '', value, flags=re.U).strip().lower()
        return re.sub(r'[-\s]+', '-', value, flags=re.U)
    else:
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
        value = re.sub(r'[^\w\s-]', '', value).strip().lower()
        return re.sub('[-\s]+', '-', value)


def urlencode(query):
    """Encode string to be used in urls (percent encoding).

    :param query: string to be encoded
    :type query: str
    :return: urlencoded string
    :rtype: str

    :Example:
        >>> urlencode('pekná líščička')
        'pekn%C3%A1%20l%C3%AD%C5%A1%C4%8Di%C4%8Dka'
    """
    if hasattr(urllib, 'parse'):
        return urllib.parse.urlencode(query)
    else:
        return urllib.urlencode(query)



"""
:mod:`zsl.interface.gearman.json_data_encoder`
----------------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

import json
import gearman
from zsl.db.model.app_model_json_encoder import AppModelJSONEncoder


class JSONDataEncoder(gearman.DataEncoder):
    @classmethod
    def encode(cls, encodable_object):
        return json.dumps(encodable_object, cls=AppModelJSONEncoder)

    @classmethod
    def decode(cls, decodable_string):
        return json.loads(decodable_string)



# -*- coding: utf-8 -*-
"""
:mod:`zsl.utils.date_helper`
----------------------------
"""
from __future__ import unicode_literals
from datetime import date, timedelta


def format_datetime_portable(ts):
    return '{0.year:{1}}-{0.month:{1}}-{0.day:{1}}T{0.hour:{1}}:{0.minute:{1}}:{0.second:{1}}'.format(ts, '02')


def format_date_portable(ts):
    return '{0.year:{1}}-{0.month:{1}}-{0.day:{1}}'.format(ts, '02')


def format_datetime_relative(dt):
    today = date.today()
    yesterday = today - timedelta(days=1)

    if dt.strftime('%Y-%m-%d') == today.strftime('%Y-%m-%d'):
        text = 'Dnes ' + dt.strftime('%H:%M')
    elif dt.strftime('%Y-%m-%d') == yesterday.strftime('%Y-%m-%d'):
        text = 'Včera ' + dt.strftime('%H:%M')
    else:
        text = dt.strftime('%d.%m.%Y %H:%M')

    return text


def format_date_relative(d):
    today = date.today()
    yesterday = today - timedelta(days=1)

    if d.strftime('%Y-%m-%d') == today.strftime('%Y-%m-%d'):
        text = 'Dnes'
    elif d.strftime('%Y-%m-%d') == yesterday.strftime('%Y-%m-%d'):
        text = 'Včera'
    else:
        text = d.strftime('%d.%m.%Y')

    return text



"""
:mod:`zsl.db.helpers.query_helper`
----------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from builtins import object
from zsl.db.helpers.pagination import Pagination
from zsl.db.helpers.sorter import Sorter
from zsl.db.helpers.query_filter import QueryFilter


class QueryHelper(object):
    def __init__(self, cls, query_filter, pagination, sorter):
        self._cls = cls

        if not isinstance(query_filter, QueryFilter):
            query_filter = QueryFilter(query_filter)
        self._query_filter = query_filter

        if not isinstance(pagination, Pagination):
            pagination = Pagination(pagination)
        self._pagination = pagination

        if not isinstance(sorter, Sorter):
            sorter = Sorter(sorter)
        self._sorter = sorter

    def execute(self, q):
        q = self._query_filter.apply_query_filter(q, self._cls)
        q = self._sorter.apply_sorter(q, self._cls)
        return self._pagination.paginate(q)

    def get_pagination(self):
        return self._pagination

    def get_sorter(self):
        return self._sorter



"""
:mod:`zsl.utils.testing`
------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import importlib
import unittest

from zsl import inject, Config
from zsl.router.method import identity_responder, set_default_responder


_test_responder_set = False


def set_test_responder():
    global _test_responder_set
    if _test_responder_set:
        return
    _test_responder_set = True
    set_default_responder(identity_responder)


@inject(config=Config)
def load_and_run_tests(config):
    """Load application test package and run it using testing util.

    :param config: application configuration, injected
    :type config: Config
    """
    set_test_responder()

    test_module = importlib.import_module(config['TEST_PACKAGE'])

    unittest.main(module=test_module)



"""
:mod:`zsl.testing.test_utils`
-----------------------------
Test utilities.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import json
from typing import Any

from zsl.task.task_data import TaskData


def parent_module(module_name):
    # type: (str) -> str
    """Return the parent module name for a module.

    :param module_name: module nam
    :type module_name: str
    :return: module's parent name
    :rtype: str

    >>> parent_module('zsl.application.module')
    'zsl.application'
    """
    return '.'.join(module_name.split('.')[:-1])


class TestTaskData(TaskData):
    """Data suitable when directly calling a task."""

    def __init__(self, payload):
        # type: (Any)->None
        super(TestTaskData, self).__init__(json.dumps(payload))



"""
:mod:`zsl.task.task_decorator`
------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>,
                  Peter Morihladko <morihladko@atteq.com>,
                  Jan Janco <janco@atteq.com>,
                  Lubos Pis <pis@atteq.com>
"""
from __future__ import unicode_literals
from builtins import *

import logging
import json

from datetime import timedelta
from typing import Callable

from flask import request
from flask.config import Config

from zsl.constants import MimeType
from zsl.task.task_data import TaskData
from zsl.db.model import AppModelJSONEncoder
from zsl.task.job_context import JobContext, WebJobContext, Responder, add_responder
import traceback
from functools import wraps
from os.path import os
from zsl.utils.file_helper import makedirs
from zsl.utils.security_helper import verify_security_data

from zsl import inject, Zsl, Injected


def log_output(f):
    """
    Logs the output value.
    """

    @wraps(f)
    def wrapper_fn(*args, **kwargs):
        res = f(*args, **kwargs)
        logging.debug("Logging result %s.", res)
        return res

    return wrapper_fn


def save_to_file(destination_filename, append=False):
    """
    Save the output value to file.
    """

    def decorator_fn(f):
        @wraps(f)
        def wrapper_fn(*args, **kwargs):
            res = f(*args, **kwargs)

            makedirs(os.path.dirname(destination_filename))
            mode = "a" if append else "w"
            with open(destination_filename, mode) as text_file:
                text_file.write(res)

            return res

        return wrapper_fn

    return decorator_fn


def json_input(f):
    """
    Expects task input data in json format and parse this data.
    """

    @wraps(f)
    def json_input_decorator(*args, **kwargs):
        # If the data is already transformed, we do not transform it any
        # further.
        task_data = _get_data_from_args(args)

        if task_data is None:
            logging.error("Task data is empty during JSON decoding.")

        if task_data.payload:
            try:
                is_transformed = request.get_json()

                # We transform the data only in the case of plain POST requests.
                if not is_transformed:
                    task_data.transform_payload(json.loads)
            except (ValueError, RuntimeError):
                logging.error(
                    "Exception while processing JSON input decorator.")
                task_data.transform_payload(json.loads)
        else:
            task_data.transform_payload(lambda _: {})

        return f(*args, **kwargs)

    return json_input_decorator


def json_output(f):
    """
    Format response to json and in case of web-request set response content type
    to 'application/json'.
    """

    @wraps(f)
    def json_output_decorator(*args, **kwargs):
        @inject(config=Config)
        def get_config(config):
            return config

        config = get_config()

        rv = f(*args, **kwargs)
        indent = None
        if config.get('DEBUG', False):
            logging.getLogger(__name__).debug("Formatting JSON nicely.")
            indent = 2
        rv = json.dumps(rv, cls=AppModelJSONEncoder, indent=indent)
        _set_json_response_content_type()
        return rv

    return json_output_decorator


def _set_json_response_content_type():
    responder = MimeSetterWebTaskResponder(MimeType.APPLICATION_JSON.value)
    add_responder(responder)


def jsonp_wrap(callback_key='callback'):
    """
    Format response to jsonp and add a callback to JSON data - a jsonp request
    """

    def decorator_fn(f):

        @wraps(f)
        def jsonp_output_decorator(*args, **kwargs):
            task_data = _get_data_from_args(args)
            data = task_data.get_data()

            if callback_key not in data:
                raise KeyError(
                    'Missing required parameter "{0}" for task.'.format(
                        callback_key))

            callback = data[callback_key]
            jsonp = f(*args, **kwargs)
            if isinstance(JobContext.get_current_context(), WebJobContext):
                JobContext.get_current_context().add_responder(
                    MimeSetterWebTaskResponder('application/javascript'))
            jsonp = "{callback}({data})".format(callback=callback, data=jsonp)

            return jsonp

        return jsonp_output_decorator

    return decorator_fn


def jsend_output(fail_exception_classes=None):
    """
    Format task result to json output in jsend specification format. See:
    http://labs.omniti.com/labs/jsend. Task return value must be dict or None.

    @param fail_exception_classes: exceptions which will produce 'fail' response
    status.
    """

    fail_exception_classes = fail_exception_classes if fail_exception_classes \
        else ()

    def decorator_fn(f):

        @wraps(f)
        @json_output
        def jsend_output_decorator(*args, **kwargs):
            try:
                rv = f(*args, **kwargs)
            except fail_exception_classes as e:
                return {'status': 'fail', 'data': {'message': str(e)}}
            except Exception as e:
                logging.error(str(e) + "\n" + traceback.format_exc())
                return {'status': 'error', 'message': 'Server error.'}

            if not isinstance(rv, dict) and rv is not None:
                msg = 'jsend_output decorator error: task must return dict ' \
                      'or None.\nTask return value: {0}.'
                logging.error(msg.format(rv))
                return {'status': 'error', 'message': 'Server error.'}

            return {'status': 'success', 'data': rv}

        return jsend_output_decorator

    return decorator_fn


def web_error_and_result(f):
    """
    Same as error_and_result decorator, except:
    If no exception was raised during task execution, ONLY IN CASE OF WEB
    REQUEST formats task result into json dictionary {'data': task return value}
    """

    @wraps(f)
    def web_error_and_result_decorator(*args, **kwargs):
        return error_and_result_decorator_inner_fn(f, True, *args, **kwargs)

    return web_error_and_result_decorator


def error_and_result(f):
    """
    Format task result into json dictionary `{'data': task return value}` if no
    exception was raised during the task execution. If there was raised an
    exception during task execution, formats task result into dictionary
    `{'error': exception message with traceback}`.
    """

    @wraps(f)
    def error_and_result_decorator(*args, **kwargs):
        return error_and_result_decorator_inner_fn(f, False, *args, **kwargs)

    return error_and_result_decorator


def error_and_result_decorator_inner_fn(f, web_only, *args, **kwargs):
    try:
        ret_val = f(*args, **kwargs)
        if web_only and not isinstance(JobContext.get_current_context(),
                                       WebJobContext):
            rv = ret_val
        else:
            rv = {'data': ret_val}
    except:
        exc = traceback.format_exc()
        logging.error(exc)
        rv = {'error': "{0}".format(exc)}

    return json.dumps(rv)


def required_data(*data):
    """
    Task decorator which checks if the given variables (indices) are stored
    inside the task data.
    """

    def decorator_fn(f):

        @wraps(f)
        def required_data_decorator(*args, **kwargs):
            task_data = _get_data_from_args(args).get_data()
            for i in data:
                if i not in task_data:
                    raise KeyError(
                        'Missing required parameter "{0}" for task.'.format(i))

            return f(*args, **kwargs)

        return required_data_decorator

    return decorator_fn


def append_get_parameters(accept_only_web=True):
    # type: (bool) -> Callable
    """
    Task decorator which appends the GET data to the task data.

    :param accept_only_web: Parameter which limits using this task only
                            with web requests.
    """

    def wrapper(f):

        @wraps(f)
        def append_get_parameters_wrapper_fn(*args, **kwargs):
            jc = JobContext.get_current_context()

            if isinstance(jc, WebJobContext):
                # Update the data with GET parameters
                web_request = jc.get_web_request()
                task_data = _get_data_from_args(args)
                data = task_data.get_data()
                data.update(web_request.args.to_dict(flat=True))
            elif accept_only_web:
                # Raise exception on non web usage if necessary
                raise Exception("append_get_parameters decorator may be used "
                                "with GET requests only.")

            return f(*args, **kwargs)

        return append_get_parameters_wrapper_fn

    return wrapper


def web_task(f):
    """
    Checks if the task is called through the web interface.
    Task return value should be in format {'data': ...}.
    """

    @wraps(f)
    def web_task_decorator(*args, **kwargs):
        jc = JobContext.get_current_context()
        if not isinstance(jc, WebJobContext):
            raise Exception(
                "The WebTask is not called through the web interface.")
        data = f(*args, **kwargs)
        jc.add_responder(WebTaskResponder(data))
        return data['data'] if 'data' in data else ""

    return web_task_decorator


def secured_task(f):
    """
    Secured task decorator.
    """

    @wraps(f)
    def secured_task_decorator(*args, **kwargs):
        task_data = _get_data_from_args(args)
        assert isinstance(task_data, TaskData)
        if not verify_security_data(task_data.get_data()['security']):
            raise SecurityException(
                task_data.get_data()['security']['hashed_token'])

        task_data.transform_payload(lambda x: x['data'])
        return f(*args, **kwargs)

    return secured_task_decorator


def xml_output(f):
    """
    Set content-type for response to WEB-REQUEST to 'text/xml'
    """

    @wraps(f)
    def xml_output_inner_fn(*args, **kwargs):
        ret_val = f(*args, **kwargs)

        if isinstance(JobContext.get_current_context(), WebJobContext):
            JobContext.get_current_context().add_responder(
                MimeSetterWebTaskResponder('text/xml'))
        return ret_val

    return xml_output_inner_fn


def file_upload(f):
    """
    Return list of `werkzeug.datastructures.FileStorage` objects - files to be
    uploaded
    """

    @wraps(f)
    def file_upload_decorator(*args, **kwargs):
        # If the data is already transformed, we do not transform it any
        # further.
        task_data = _get_data_from_args(args)

        if task_data is None:
            logging.error("Task data is empty during FilesUploadDecorator.")

        task_data.transform_payload(lambda _: request.files.getlist('file'))
        return f(*args, **kwargs)

    return file_upload_decorator


def _get_data_from_args(args):
    task_data = None
    for d in args:
        if isinstance(d, TaskData):
            task_data = d

    return task_data


class SecurityException(Exception):
    def __init__(self, hashed_token):
        Exception.__init__(self,
                           "Invalid hashed token '{0}'.".format(hashed_token))
        self._hashed_token = hashed_token

    def get_hashed_token(self):
        return self._hashed_token


class WebTaskResponder(Responder):
    def __init__(self, data):
        self.data = data

    def respond(self, response):
        for k in self.data:
            if k == 'headers':
                for header_name in self.data[k]:
                    response.headers[header_name] = self.data[k][header_name]
            else:
                setattr(response, k, self.data[k])


class MimeSetterWebTaskResponder(Responder):
    def __init__(self, mime):
        self._mime = mime

    def respond(self, r):
        r.content_type = self._mime


class CrossdomainWebTaskResponder(Responder):
    """
    source: http://flask.pocoo.org/snippets/56/
    """

    @inject(app=Zsl, config=Config)
    def __init__(self, origin=None, methods=None, headers=None,
                 max_age=21600, app=Injected, config=Injected):
        self._app = app

        if methods is not None:
            methods = ', '.join(sorted(x.upper() for x in methods))
        self.methods = methods

        if headers is not None and not isinstance(headers, (str, bytes)):
            headers = ', '.join(x.upper() for x in headers)
        self.headers = headers

        if origin is None:
            origin = config.get('CORS', {}).get('origin', '')

        if not isinstance(origin, (str, bytes)):
            origin = ', '.join(origin)
        self.origin = origin

        if isinstance(max_age, timedelta):
            max_age = max_age.total_seconds()
        self.max_age = max_age

    def get_methods(self):
        if self.methods is not None:
            return self.methods

        options_resp = self._app.make_default_options_response()
        return options_resp.headers['allow']

    def respond(self, response):
        headers = response.headers

        headers['Access-Control-Allow-Origin'] = self.origin
        headers['Access-Control-Allow-Methods'] = self.get_methods()
        headers['Access-Control-Max-Age'] = str(self.max_age)
        if self.headers is not None:
            headers['Access-Control-Allow-Headers'] = self.headers
        else:
            response.headers['Access-Control-Allow-Headers'] = \
                'accept, origin, content-type, authorization'

        response.headers['Access-Control-Expose-Headers'] = \
            'location, x-total-count, link'

        return response


def crossdomain(origin=None, methods=None, headers=None, max_age=21600):
    def decorator(f):
        responder = CrossdomainWebTaskResponder(origin, methods, headers,
                                                max_age)

        @wraps(f)
        def crossdomain_inner_fn(*args, **kwargs):
            rv = f(*args, **kwargs)
            add_responder(responder)
            return rv

        return crossdomain_inner_fn

    return decorator



"""
:mod:`zsl.interface.cli`
------------------------

.. moduleauthor:: Peter Morihladko <peter@atteq.com>, Martin Babka <babka@atteq.com>

The module is responsible for creating the CLI interface of ZSL and provides the main `click` group for all the CLI 
commands and groups. If any command is to be added add it to this group `cli` defined in this module.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import click
import sys

from zsl import __version__

click.disable_unicode_literals_warning = True


def _get_version(ctx, _, value):
    """Click callback for option to show current ZSL version."""

    if not value or ctx.resilient_parsing:
        return
    message = 'Zsl %(version)s\nPython %(python_version)s'
    click.echo(message % {
        'version': __version__,
        'python_version': sys.version,
    }, color=ctx.color)
    ctx.exit()


@click.group(help="ZSL CLI toolkit")
@click.option('--version', help="Show ZSL version", expose_value=False,
              is_flag=True, is_eager=True, callback=_get_version)
def cli():
    """The command group for ZSL."""
    pass


@cli.command(help="Open interactive shell (bpython) with the ZSL environment",
             context_settings=dict(ignore_unknown_options=True))
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def shell(args):
    import bpython
    bpython.embed(args=list(args))



"""
:mod:`zsl.resource.resource_helper`
-----------------------------------

.. moduleauthor:: Peter Morihladko
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from future.utils import viewitems, viewvalues

from sqlalchemy.orm import class_mapper, joinedload
from sqlalchemy import desc, asc, and_


def filter_from_url_arg(model_cls, query, arg, query_operator=and_,
                        arg_types=None):
    """
    Parse filter URL argument ``arg`` and apply to ``query``

    Example: 'column1<=value,column2==value' -> query.filter(Model.column1 <= value, Model.column2 == value)
    """

    fields = arg.split(',')
    mapper = class_mapper(model_cls)

    if not arg_types:
        arg_types = {}

    exprs = []
    joins = set()
    for expr in fields:
        if expr == "":
            continue

        e_mapper = mapper
        e_model_cls = model_cls

        operator = None
        method = None
        for op in operator_order:
            if op in expr:
                operator = op
                method = operator_to_method[op]
                break

        if operator is None:
            raise Exception('No operator in expression "{0}".'.format(expr))

        (column_names, value) = expr.split(operator)

        column_names = column_names.split('__')
        value = value.strip()

        for column_name in column_names:
            if column_name in arg_types:
                typed_value = arg_types[column_name](value)
            else:
                typed_value = value

            if column_name in e_mapper.relationships:
                joins.add(column_name)
                e_model_cls = e_mapper.attrs[column_name].mapper.class_
                e_mapper = class_mapper(e_model_cls)

        if hasattr(e_model_cls, column_name):
            column = getattr(e_model_cls, column_name)
            exprs.append(getattr(column, method)(typed_value))
        else:
            raise Exception('Invalid property {0} in class {1}.'.format(column_name, e_model_cls))

    return query.join(*joins).filter(query_operator(*exprs))


operator_to_method = {
    '::like::': 'like',
    '==': '__eq__',
    '<=': '__le__',
    '>=': '__ge__',
    '!=': '__ne__',
    '<': '__lt__',
    '>': '__gt__'
}
operator_order = ['::like::', '==', '<=', '>=', '!=', '<', '>']


def order_from_url_arg(model_cls, query, arg):
    fields = arg.split(',')
    mapper = class_mapper(model_cls)

    orderings = []
    joins = []
    for field in fields:
        e_mapper = mapper
        e_model_cls = model_cls

        if field[0] == '-':
            column_names = field[1:]
            direction = 'desc'
        else:
            column_names = field
            direction = 'asc'

        column_names = column_names.split('__')

        for column_name in column_names:
            if column_name in e_mapper.relationships:
                joins.append(column_name)
                e_model_cls = e_mapper.attrs[column_name].mapper.class_
                e_mapper = class_mapper(e_model_cls)

        if hasattr(e_model_cls, column_name):
            column = getattr(e_model_cls, column_name)
            order_by = asc(column) if direction == 'asc' else desc(column)
            orderings.append(order_by)
        else:
            raise Exception('Invalid property {0} in class {1}.'.format(column_name, model_cls))

    return query.join(*joins).order_by(*orderings)


def create_related_tree(fields):
    tree = {}
    for field in fields:
        fs = field.split('__')
        node = tree

        for f in fs:
            if f not in node:
                node[f] = {}
            node = node[f]

    # replace empty {} with None, to represent leafs
    q = [tree]
    while len(q) > 0:
        node = q.pop()

        for k, v in viewitems(node):
            if len(v) > 0:
                q.append(v)
            else:
                node[k] = None

    return tree


def model_tree(name, model_cls, visited=None):
    """Create a simple tree of model's properties and its related models.
    
    It traverse trough relations, but ignore any loops.
    
    :param name: name of the model 
    :type name: str
    :param model_cls: model class
    :param visited: set of visited models
    :type visited: list or None
    :return: a dictionary where values are lists of string or other \
    dictionaries
    """
    if not visited:
        visited = set()

    visited.add(model_cls)

    mapper = class_mapper(model_cls)
    columns = [column.key for column in mapper.column_attrs]
    related = [model_tree(rel.key, rel.mapper.entity, visited)
               for rel in mapper.relationships if rel.mapper.entity not in visited]

    return {name: columns + related}


def flat_model(tree):
    """Flatten the tree into a list of properties adding parents as prefixes."""
    names = []
    for columns in viewvalues(tree):
        for col in columns:
            if isinstance(col, dict):
                col_name = list(col)[0]
                names += [col_name + '__' + c for c in flat_model(col)]
            else:
                names.append(col)

    return names


def apply_related(model_cls, query, related_fields):
    """

    """
    mapper = class_mapper(model_cls)

    loads = []

    # TODO zatial iba pre
    for field in related_fields:
        if field in mapper.relationships:
            loads.append(joinedload(getattr(model_cls, field)))

    if loads:
        query = query.options(*loads)

    return query


def related_from_fields(fields):
    return [field.rsplit('__', 1)[0] for field in fields if '__' in field]



"""
:mod:`zsl.tasks.asl.schedule_gearman_task`
------------------------------------------
"""
from __future__ import unicode_literals
from builtins import object
from zsl.utils.gearman_helper import schedule_gearman_task


class ScheduleGearmanTask(object):

    def perform(self, data):
        data = data.get_data()

        # Create gearman.
        schedule_gearman_task(data['path'], data['data'])

        return 'job submitted'



"""
:mod:`zsl.tasks.asl.kill_worker_task`
-------------------------------------

Created on 22.12.2012

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

from builtins import object
from zsl import Zsl
from injector import inject
from zsl.task.job_context import JobContext, WebJobContext
from zsl.interface.task_queue import KillWorkerException


class KillWorkerTask(object):
    @inject(app=Zsl)
    def __init__(self, app):
        self._app = app

    @staticmethod
    def perform(data):
        if isinstance(JobContext.get_current_context(), WebJobContext):
            raise Exception("Can not kill worker from web!")

        raise KillWorkerException()



"""
:mod:`zsl.utils.model_helper`
-----------------------------

Helper module for working with models.
"""

from __future__ import unicode_literals

import logging


def update_model(raw_model, app_model, forbidden_keys=None, inverse=False):
    """Updates the `raw_model` according to the values in the `app_model`.

    :param raw_model: Raw model which gets updated.
    :param app_model: App model holding the data.
    :param forbidden_keys: Data/attributes which will not be updated.
    :type forbidden_keys: list
    :param inverse: If the value is `True` all `app_model` attributes which are contained in the `raw_model` are
                    updated. If the value is `False` all `raw_model` properties which are in the `app_model` will be
                    updated.
    """
    if forbidden_keys is None:
        forbidden_keys = []

    if type(app_model) != dict:
        app_model = app_model.__dict__

    if inverse:
        for k in app_model:
            logging.debug("Considering property {0}.".format(k))
            if (hasattr(raw_model, k)) and (k not in forbidden_keys):
                logging.debug("Setting property {0} to value '{1}'.".format(k, app_model[k]))
                setattr(raw_model, k, app_model[k])
    else:
        for k in raw_model.__dict__:
            logging.debug("Considering property {0}.".format(k))
            if (k in app_model) and (k not in forbidden_keys):
                logging.debug("Setting property {0} to value '{1}'.".format(k, app_model[k]))
                setattr(raw_model, k, app_model[k])



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *


def extract_data(request):
    data = request.form.to_dict(flat=True)
    json_data = request.get_json()
    return json_data if json_data else data



"""
:mod:`zsl.application.modules.task_router_module`
-------------------------------------------------
"""
from __future__ import unicode_literals

from injector import Module, singleton, ClassProvider

from zsl.router.task import TaskRouter


class TaskRouterModule(Module):
    """Adds task router to current configuration."""
    task_provider = ClassProvider(TaskRouter)

    def configure(self, binder):
        binder.bind(TaskRouter, to=self.task_provider, scope=singleton)



"""
:mod:`zsl.cache.cache_module`
-----------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from future.utils import with_metaclass

from builtins import object
import abc


class CacheModule(with_metaclass(abc.ABCMeta, object)):
    """
    Cache abstraction layer - module for caching.
    """

    @abc.abstractmethod
    def set_key(self, key, value, timeout):
        """
        Saves the ``key``-``value`` pair for ``timeout`` seconds.
        """
        pass

    @abc.abstractmethod
    def invalidate_key(self, key):
        """
        Deletes the ``key``.
        """
        pass

    @abc.abstractmethod
    def contains_key(self, key):
        """
        Check if the ``key`` is contained in the cache.
        """
        pass

    @abc.abstractmethod
    def get_key(self, key):
        """
        Returns the value associated with the ``key``.

        TODO this should be named get_by_key or just simply get(key), get_key sounds like it would return a key
        """
        pass

    @abc.abstractmethod
    def contains_list(self, key):
        """
        Check if the ``key`` is in the cache.
        """
        pass

    @abc.abstractmethod
    def get_list(self, key):
        """
        Returns the list associated with the ``key``.
        """
        pass

    @abc.abstractmethod
    def invalidate_by_glob(self, glob):
        """
        Invalidates the keys by given glob.

        :param string glob: All the keys matching the given glob will be invalidated.
        """
